# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = ["ElasticLBDriver"]


from libcloud.utils.py3 import httplib
from libcloud.utils.xml import findall, findtext
from libcloud.common.aws import AWSGenericResponse, SignedAWSConnection
from libcloud.loadbalancer.base import Driver, Member, LoadBalancer
from libcloud.loadbalancer.types import State

VERSION = "2012-06-01"
HOST = "elasticloadbalancing.%s.amazonaws.com"
ROOT = "/%s/" % (VERSION)
NS = "http://elasticloadbalancing.amazonaws.com/doc/{}/".format(VERSION)


class ELBResponse(AWSGenericResponse):
    """
    Amazon ELB response class.
    """

    namespace = NS
    exceptions = {}
    xpath = "Error"


class ELBConnection(SignedAWSConnection):
    version = VERSION
    host = HOST
    responseCls = ELBResponse
    service_name = "elasticloadbalancing"


class ElasticLBDriver(Driver):
    name = "Amazon Elastic Load Balancing"
    website = "http://aws.amazon.com/elasticloadbalancing/"
    connectionCls = ELBConnection
    signature_version = "4"

    def __init__(self, access_id, secret, region, token=None):
        self.token = token
        self.region = region
        self.region_name = region
        super().__init__(access_id, secret, token=token, host=HOST % region, region=region)

    def list_protocols(self):
        return ["tcp", "ssl", "http", "https"]

    def list_balancers(self, ex_fetch_tags=False):
        params = {"Action": "DescribeLoadBalancers"}
        data = self.connection.request(ROOT, params=params).object
        balancers = self._to_balancers(data)

        if ex_fetch_tags:
            for balancer in balancers:
                self._ex_populate_balancer_tags(balancer)

        return balancers

    def create_balancer(
        self,
        name,
        port,
        protocol,
        algorithm,
        members,
        ex_members_availability_zones=None,
    ):
        if ex_members_availability_zones is None:
            ex_members_availability_zones = ["a"]

        params = {
            "Action": "CreateLoadBalancer",
            "LoadBalancerName": name,
            "Listeners.member.1.InstancePort": str(port),
            "Listeners.member.1.InstanceProtocol": protocol.upper(),
            "Listeners.member.1.LoadBalancerPort": str(port),
            "Listeners.member.1.Protocol": protocol.upper(),
        }

        for i, z in enumerate(ex_members_availability_zones):
            zone = "".join((self.region, z))
            params["AvailabilityZones.member.%d" % (i + 1)] = zone

        data = self.connection.request(ROOT, params=params).object

        balancer = LoadBalancer(
            id=name,
            name=name,
            state=State.PENDING,
            ip=findtext(element=data, xpath="DNSName", namespace=NS),
            port=port,
            driver=self.connection.driver,
        )
        balancer._members = []
        return balancer

    def destroy_balancer(self, balancer):
        params = {"Action": "DeleteLoadBalancer", "LoadBalancerName": balancer.id}
        self.connection.request(ROOT, params=params)
        return True

    def get_balancer(self, balancer_id, ex_fetch_tags=False):
        params = {
            "Action": "DescribeLoadBalancers",
            "LoadBalancerNames.member.1": balancer_id,
        }
        data = self.connection.request(ROOT, params=params).object
        balancer = self._to_balancers(data)[0]

        if ex_fetch_tags:
            balancer = self._ex_populate_balancer_tags(balancer)

        return balancer

    def balancer_attach_compute_node(self, balancer, node):
        params = {
            "Action": "RegisterInstancesWithLoadBalancer",
            "LoadBalancerName": balancer.id,
            "Instances.member.1.InstanceId": node.id,
        }
        self.connection.request(ROOT, params=params)
        balancer._members.append(Member(node.id, None, None, balancer=self))

    def balancer_detach_member(self, balancer, member):
        params = {
            "Action": "DeregisterInstancesFromLoadBalancer",
            "LoadBalancerName": balancer.id,
            "Instances.member.1.InstanceId": member.id,
        }
        self.connection.request(ROOT, params=params)
        balancer._members = [m for m in balancer._members if m.id != member.id]
        return True

    def balancer_list_members(self, balancer):
        return balancer._members

    def ex_list_balancer_policies(self, balancer):
        """
        Return a list of policy description string.

        :rtype: ``list`` of ``str``
        """
        params = {
            "Action": "DescribeLoadBalancerPolicies",
            "LoadBalancerName": balancer.id,
        }

        data = self.connection.request(ROOT, params=params).object
        return self._to_policies(data)

    def ex_list_balancer_policy_types(self):
        """
        Return a list of policy type description string.

        :rtype: ``list`` of ``str``
        """
        params = {"Action": "DescribeLoadBalancerPolicyTypes"}

        data = self.connection.request(ROOT, params=params).object
        return self._to_policy_types(data)

    def ex_create_balancer_policy(self, name, policy_name, policy_type, policy_attributes=None):
        """
        Create a new load balancer policy

        :param name: Balancer name to create the policy for
        :type  name: ``str``

        :param policy_name: policy to be created
        :type  policy_name: ``str``

        :param policy_type: policy type being used to create policy.
        :type  policy_type: ``str``

        :param policy_attributes: Each list contain values, ['AttributeName',
                                                             'value']
        :type  policy_attributes: ``PolicyAttribute list``
        """
        params = {
            "Action": "CreateLoadBalancerPolicy",
            "LoadBalancerName": name,
            "PolicyName": policy_name,
            "PolicyTypeName": policy_type,
        }

        if policy_attributes is not None:
            for index, (name, value) in enumerate(policy_attributes.iteritems(), 1):
                params["PolicyAttributes.member.%d. \
                        AttributeName" % (index)] = name
                params["PolicyAttributes.member.%d. \
                        AttributeValue" % (index)] = value

        response = self.connection.request(ROOT, params=params)
        return response.status == httplib.OK

    def ex_delete_balancer_policy(self, name, policy_name):
        """
        Delete a load balancer policy

        :param name: balancer name for which policy will be deleted
        :type  name: ``str``

        :param policy_name: The Mnemonic name for the policy being deleted
        :type  policy_name: ``str``
        """
        params = {
            "Action": "DeleteLoadBalancerPolicy",
            "LoadBalancerName": name,
            "PolicyName": policy_name,
        }

        response = self.connection.request(ROOT, params=params)
        return response.status == httplib.OK

    def ex_set_balancer_policies_listener(self, name, port, policies):
        """
        Associates, updates, or disables a policy with a listener on
        the load balancer

        :param name: balancer name to set policies for listerner
        :type  name: ``str``

        :param port: port to use
        :type  port: ``str``

        :param policies: List of policies to be associated with the balancer
        :type  policies: ``string list``
        """
        params = {
            "Action": "SetLoadBalancerPoliciesOfListener",
            "LoadBalancerName": name,
            "LoadBalancerPort": str(port),
        }

        if policies:
            params = self._create_list_params(params, policies, "PolicyNames.member.%d")

        response = self.connection.request(ROOT, params=params)
        return response.status == httplib.OK

    def ex_set_balancer_policies_backend_server(self, name, instance_port, policies):
        """
        Replaces the current set of policies associated with a port on
        which the back-end server is listening with a new set of policies

        :param name: balancer name to set policies of backend server
        :type  name: ``str``

        :param instance_port: Instance Port
        :type  instance_port: ``int``

        :param policies: List of policies to be associated with the balancer
        :type  policies: ``string list`
        """
        params = {
            "Action": "SetLoadBalancerPoliciesForBackendServer",
            "LoadBalancerName": name,
            "InstancePort": str(instance_port),
        }

        if policies:
            params = self._create_list_params(params, policies, "PolicyNames.member.%d")

        response = self.connection.request(ROOT, params=params)
        return response.status == httplib.OK

    def ex_create_balancer_listeners(self, name, listeners=None):
        """
        Creates one or more listeners on a load balancer for the specified port

        :param name: The mnemonic name associated with the load balancer
        :type  name: ``str``

        :param listeners: Each tuple contain values, (LoadBalancerPortNumber,
                          InstancePortNumber, Protocol,[SSLCertificateId])
        :type  listeners: ``list of tuple`
        """
        params = {"Action": "CreateLoadBalancerListeners", "LoadBalancerName": name}

        for index, listener in enumerate(listeners):
            i = index + 1
            protocol = listener[2].upper()
            params["Listeners.member.%d.LoadBalancerPort" % i] = listener[0]
            params["Listeners.member.%d.InstancePort" % i] = listener[1]
            params["Listeners.member.%d.Protocol" % i] = listener[2]
            if protocol == "HTTPS" or protocol == "SSL":
                params["Listeners.member.%d.   \
                        SSLCertificateId" % i] = listener[3]
        else:
            return False

        response = self.connection.request(ROOT, params=params)
        return response.status == httplib.OK

    def _to_policies(self, data):
        xpath = "DescribeLoadBalancerPoliciesResult/PolicyDescriptions/member"
        return [
            findtext(element=el, xpath="PolicyName", namespace=NS)
            for el in findall(element=data, xpath=xpath, namespace=NS)
        ]

    def _to_policy_types(self, data):
        xpath = "DescribeLoadBalancerPolicyTypesResult/"
        xpath += "PolicyTypeDescriptions/member"
        return [
            findtext(element=el, xpath="PolicyTypeName", namespace=NS)
            for el in findall(element=data, xpath=xpath, namespace=NS)
        ]

    def _to_balancers(self, data):
        xpath = "DescribeLoadBalancersResult/LoadBalancerDescriptions/member"
        return [self._to_balancer(el) for el in findall(element=data, xpath=xpath, namespace=NS)]

    def _to_balancer(self, el):
        name = findtext(element=el, xpath="LoadBalancerName", namespace=NS)
        dns_name = findtext(el, xpath="DNSName", namespace=NS)
        port = findtext(el, xpath="LoadBalancerPort", namespace=NS)

        balancer = LoadBalancer(
            id=name,
            name=name,
            state=State.UNKNOWN,
            ip=dns_name,
            port=port,
            driver=self.connection.driver,
        )

        xpath = "Instances/member/InstanceId"
        members = findall(element=el, xpath=xpath, namespace=NS)
        balancer._members = []

        for m in members:
            balancer._members.append(Member(m.text, None, None, balancer=balancer))

        return balancer

    def _to_tags(self, data):
        """
        return tags dict
        """
        tags = {}
        xpath = "DescribeTagsResult/TagDescriptions/member/Tags/member"
        for el in findall(element=data, xpath=xpath, namespace=NS):
            key = findtext(element=el, xpath="Key", namespace=NS)
            value = findtext(element=el, xpath="Value", namespace=NS)
            if key:
                tags[key] = value

        return tags

    def _create_list_params(self, params, items, label):
        """
        return parameter list
        """
        if isinstance(items, str):
            items = [items]
        for index, item in enumerate(items):
            params[label % (index + 1)] = item
        return params

    def _ex_connection_class_kwargs(self):
        kwargs = super()._ex_connection_class_kwargs()
        if hasattr(self, "token") and self.token is not None:
            kwargs["token"] = self.token
            kwargs["signature_version"] = "4"
        else:
            kwargs["signature_version"] = self.signature_version

        return kwargs

    def _ex_list_balancer_tags(self, balancer_id):
        params = {"Action": "DescribeTags", "LoadBalancerNames.member.1": balancer_id}
        data = self.connection.request(ROOT, params=params).object
        return self._to_tags(data)

    def _ex_populate_balancer_tags(self, balancer):
        tags = balancer.extra.get("tags", {})
        tags.update(self._ex_list_balancer_tags(balancer.id))
        if tags:
            balancer.extra["tags"] = tags

        return balancer



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import ssl
import time
import socket
import logging
from datetime import datetime, timedelta
from functools import wraps

from libcloud.utils.py3 import httplib
from libcloud.common.exceptions import RateLimitReachedError

__all__ = [
    "Retry",
    "RetryForeverOnRateLimitError",
]

_logger = logging.getLogger(__name__)
# Error message which indicates a transient SSL error upon which request
# can be retried
TRANSIENT_SSL_ERROR = "The read operation timed out"


class TransientSSLError(ssl.SSLError):
    """Represent transient SSL errors, e.g. timeouts"""

    pass


# Constants used by the ``retry`` class
# All the time values (timeout, delay, backoff) are in seconds
DEFAULT_TIMEOUT = 30  # default retry timeout
DEFAULT_DELAY = 1  # default sleep delay used in each iterator
DEFAULT_BACKOFF = 1  # retry backup multiplier
RETRY_EXCEPTIONS = (
    RateLimitReachedError,
    socket.error,
    socket.gaierror,
    httplib.NotConnected,
    httplib.ImproperConnectionState,
    TransientSSLError,
)


class MinimalRetry:
    def __init__(
        self,
        retry_delay=DEFAULT_DELAY,
        timeout=DEFAULT_TIMEOUT,
        backoff=DEFAULT_BACKOFF,
    ):
        """
        Wrapper around retrying that helps to handle common transient
        exceptions.

        This minimalistic version only retries SSL errors and rate limiting.

        :param retry_delay: retry delay between the attempts.
        :param timeout: maximum time to wait.
        :param backoff: multiplier added to delay between attempts.

        :Example:

        retry_request = MinimalRetry(timeout=1, retry_delay=1, backoff=1)
        retry_request(self.connection.request)()
        """

        if retry_delay is None:
            retry_delay = DEFAULT_DELAY
        if timeout is None:
            timeout = DEFAULT_TIMEOUT
        if backoff is None:
            backoff = DEFAULT_BACKOFF

        timeout = max(timeout, 0)

        self.retry_delay = retry_delay
        self.timeout = timeout
        self.backoff = backoff

    def __call__(self, func):
        def transform_ssl_error(function, *args, **kwargs):
            try:
                return function(*args, **kwargs)
            except ssl.SSLError as exc:
                if TRANSIENT_SSL_ERROR in str(exc):
                    raise TransientSSLError(*exc.args)

                raise exc

        @wraps(func)
        def retry_loop(*args, **kwargs):
            current_delay = self.retry_delay
            end = datetime.now() + timedelta(seconds=self.timeout)
            last_exc = None

            while datetime.now() < end:
                try:
                    return transform_ssl_error(func, *args, **kwargs)
                except Exception as exc:
                    last_exc = exc

                    if isinstance(exc, RateLimitReachedError):
                        _logger.debug("You are being rate limited, backing " "off...")

                        # NOTE: Retry after defaults to 0 in the
                        # RateLimitReachedError class so we a use more
                        # reasonable default in case that attribute is not
                        # present. This way we prevent busy waiting, etc.
                        retry_after = exc.retry_after if exc.retry_after else 2
                        time.sleep(retry_after)

                        # Reset delay if we're told to wait due to rate
                        # limiting
                        current_delay = self.retry_delay
                    elif self.should_retry(exc):
                        time.sleep(current_delay)
                        current_delay *= self.backoff
                    else:
                        raise

            raise last_exc

        return retry_loop

    def should_retry(self, exception):
        return False


class Retry(MinimalRetry):
    def __init__(
        self,
        retry_exceptions=RETRY_EXCEPTIONS,
        retry_delay=DEFAULT_DELAY,
        timeout=DEFAULT_TIMEOUT,
        backoff=DEFAULT_BACKOFF,
    ):
        """
        Wrapper around retrying that helps to handle common transient
        exceptions.

        This version retries the errors that
        `libcloud.utils.retry:MinimalRetry` retries and all errors of the
        exception types that are given.

        :param retry_exceptions: types of exceptions to retry on.
        :param retry_delay: retry delay between the attempts.
        :param timeout: maximum time to wait.
        :param backoff: multiplier added to delay between attempts.

        :Example:

        retry_request = Retry(retry_exceptions=(httplib.NotConnected,),
                              timeout=1, retry_delay=1, backoff=1)
        retry_request(self.connection.request)()
        """

        super().__init__(retry_delay=retry_delay, timeout=timeout, backoff=backoff)
        if retry_exceptions is None:
            retry_exceptions = RETRY_EXCEPTIONS
        self.retry_exceptions = retry_exceptions

    def should_retry(self, exception):
        return isinstance(exception, tuple(self.retry_exceptions))


class RetryForeverOnRateLimitError(Retry):
    """
    This class is only here for backward compatibility reasons with
    pre-Libcloud v3.3.2.

    If works by ignoring timeout argument and retrying forever until API
    is returning 429 RateLimitReached errors.

    In most cases using this class is not a good idea since it can cause code
    to hang and retry for ever in case API continues to return retry limit
    reached.
    """

    def __call__(self, func):
        def transform_ssl_error(function, *args, **kwargs):
            try:
                return function(*args, **kwargs)
            except ssl.SSLError as exc:
                if TRANSIENT_SSL_ERROR in str(exc):
                    raise TransientSSLError(*exc.args)

                raise exc

        @wraps(func)
        def retry_loop(*args, **kwargs):
            current_delay = self.retry_delay
            end = datetime.now() + timedelta(seconds=self.timeout)

            while True:
                try:
                    return transform_ssl_error(func, *args, **kwargs)
                except Exception as exc:
                    if isinstance(exc, RateLimitReachedError):
                        time.sleep(exc.retry_after)

                        # Reset retries if we're told to wait due to rate
                        # limiting
                        current_delay = self.retry_delay
                        end = datetime.now() + timedelta(seconds=exc.retry_after + self.timeout)
                    elif datetime.now() >= end:
                        raise
                    elif self.should_retry(exc):
                        time.sleep(current_delay)
                        current_delay *= self.backoff
                    else:
                        raise

        return retry_loop



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Equinix Metal Driver
"""

try:  # Try to use asyncio to perform requests in parallel across projects
    import asyncio
except ImportError:  # If not available will do things serially
    asyncio = None

import json

from libcloud.utils.py3 import httplib
from libcloud.common.base import JsonResponse, ConnectionKey
from libcloud.compute.base import Node, KeyPair, NodeSize, NodeImage, NodeDriver, NodeLocation
from libcloud.compute.types import Provider, NodeState, InvalidCredsError

EQUINIXMETAL_ENDPOINT = "api.equinix.com"

# True to use async io if available (aka running under Python 3)
USE_ASYNC_IO_IF_AVAILABLE = True


def use_asyncio():
    return asyncio is not None and USE_ASYNC_IO_IF_AVAILABLE


class EquinixMetalResponse(JsonResponse):
    valid_response_codes = [
        httplib.OK,
        httplib.ACCEPTED,
        httplib.CREATED,
        httplib.NO_CONTENT,
    ]

    def parse_error(self):
        if self.status == httplib.UNAUTHORIZED:
            body = self.parse_body()
            raise InvalidCredsError(body.get("error"))
        else:
            body = self.parse_body()

            if "message" in body:
                error = "{} (code: {})".format(body.get("message"), self.status)
            elif "errors" in body:
                error = body.get("errors")
            else:
                error = body
            raise Exception(error)

    def success(self):
        return self.status in self.valid_response_codes


class EquinixMetalConnection(ConnectionKey):
    """
    Connection class for the Equinix Metal driver.
    """

    host = EQUINIXMETAL_ENDPOINT
    responseCls = EquinixMetalResponse

    def add_default_headers(self, headers):
        """
        Add headers that are necessary for every request
        """
        headers["Content-Type"] = "application/json"
        headers["X-Auth-Token"] = self.key
        headers["X-Consumer-Token"] = (
            "kcrhMn7hwG8Ceo2hAhGFa2qpxLBvVHxEjS9ue8iqmsNkeeB2iQgMq4dNc1893pYu"
        )

        return headers


class EquinixMetalNodeDriver(NodeDriver):
    """
    Equinix Metal NodeDriver
    """

    connectionCls = EquinixMetalConnection
    type = Provider.EQUINIXMETAL
    name = "EquinixMetal"
    website = "https://metal.equinix.com/"

    NODE_STATE_MAP = {
        "queued": NodeState.PENDING,
        "provisioning": NodeState.PENDING,
        "rebuilding": NodeState.PENDING,
        "powering_on": NodeState.REBOOTING,
        "powering_off": NodeState.REBOOTING,
        "rebooting": NodeState.REBOOTING,
        "inactive": NodeState.STOPPED,
        "deleted": NodeState.TERMINATED,
        "deprovisioning": NodeState.TERMINATED,
        "failed": NodeState.ERROR,
        "active": NodeState.RUNNING,
    }

    def __init__(self, key, project=None):
        """
        Initialize a NodeDriver for Equinix Metal using the API token
        and optionally the project (name or id).

        If project name is specified we validate it lazily and populate
        self.project_id during the first access of self.projects variable
        """
        super().__init__(key=key)

        self.project_name = project
        self.project_id = None

        # Lazily populated on first access to self.project
        self._project = project

        # Variable which indicates if self._projects has been populated yet and
        # has been called self._project validated
        self._projects_populated = False
        self._projects = None

    @property
    def projects(self):
        """
        Lazily retrieve projects and set self.project_id variable on initial
        access to self.projects variable.
        """

        if not self._projects_populated:
            # NOTE: Each EquinixMetal account needs at least one project,
            # but to be on the safe side and avoid infinite loop
            # in case there are no projects on the account, we don't use
            # a more robust way to determine
            # if project list has been populated yet
            self._projects = self.ex_list_projects()
            self._projects_populated = True

            # If project name is specified, verify it's valid and populate
            # self.project_id

            if self._project:
                for project_obj in self._projects:
                    if self._project in [project_obj.name, project_obj.id]:
                        self.project_id = project_obj.id

                        break

                if not self.project_id:
                    # Invalid project name
                    self.project_name = None

        return self._projects

    def ex_list_projects(self):
        projects = []
        data = self.connection.request("/metal/v1/projects").object
        projects = data.get("projects")

        if projects:
            projects = [Project(project) for project in projects]

        return projects

    def list_nodes(self, ex_project_id=None):
        if ex_project_id:
            return self.ex_list_nodes_for_project(ex_project_id=ex_project_id)

        # if project has been specified during driver initialization, then
        # return nodes for this project only

        if self.project_id:
            return self.ex_list_nodes_for_project(ex_project_id=self.project_id)

        # In case of Python2 perform requests serially

        if not use_asyncio():
            nodes = []

            for project in self.projects:
                nodes.extend(self.ex_list_nodes_for_project(ex_project_id=project.id))

            return nodes
        # In case of Python3 use asyncio to perform requests in parallel

        return self.list_resources_async("nodes")

    def list_resources_async(self, resource_type):
        # The _list_nodes function is defined dynamically using exec in
        # order to prevent a SyntaxError in Python2 due to "yield from".
        # This cruft can be removed once Python2 support is no longer
        # required.
        assert resource_type in ["nodes"]
        glob = globals()
        loc = locals()
        exec(
            """
import asyncio
@asyncio.coroutine
def _list_async(driver):
    projects = [project.id for project in driver.projects]
    loop = asyncio.get_event_loop()
    futures = [
        loop.run_in_executor(None, driver.ex_list_%s_for_project, p)
        for p in projects
    ]
    retval = []
    for future in futures:
        result = yield from future
        retval.extend(result)

    return retval""" % resource_type,
            glob,
            loc,
        )
        try:
            loop = asyncio.get_event_loop()
        except RuntimeError:
            asyncio.set_event_loop(asyncio.new_event_loop())
            loop = asyncio.get_event_loop()

        return loop.run_until_complete(loc["_list_async"](loc["self"]))

    def ex_list_nodes_for_project(self, ex_project_id, include="plan", page=1, per_page=1000):
        params = {"include": include, "page": page, "per_page": per_page}
        data = self.connection.request(
            "/metal/v1/projects/%s/devices" % (ex_project_id), params=params
        ).object["devices"]

        return list(map(self._to_node, data))

    def list_locations(self):
        data = self.connection.request("/metal/v1/locations/metros").object["metros"]

        return list(map(self._to_location, data))

    def list_images(self):
        data = self.connection.request("/metal/v1/operating-systems").object["operating_systems"]

        return list(map(self._to_image, data))

    def list_sizes(self, ex_project_id=None):
        project_id = (
            ex_project_id or self.project_id or (len(self.projects) and self.projects[0].id)
        )

        if project_id:
            data = self.connection.request("/metal/v1/projects/%s/plans" % project_id).object[
                "plans"
            ]
        else:  # This only works with personal tokens
            data = self.connection.request("/metal/v1/plans").object["plans"]

        return [self._to_size(size) for size in data if size.get("line") == "baremetal"]

    def create_node(
        self,
        name,
        size,
        image,
        location,
        ex_project_id=None,
        ip_addresses=None,
        cloud_init=None,
        **kwargs,
    ):
        """
        Create a node.

        :return: The newly created node.
        :rtype: :class:`Node`
        """
        # if project has been specified on initialization of driver, then
        # create on this project

        if self.project_id:
            ex_project_id = self.project_id
        else:
            if not ex_project_id:
                raise Exception("ex_project_id needs to be specified")

        location_code = location.extra["code"]

        if not self._valid_location:
            raise ValueError(
                "Failed to create node: valid parameter metro [code] is required in the input"
            )

        params = {
            "hostname": name,
            "plan": size.id,
            "operating_system": image.id,
            "metro": location_code,
            "include": "plan",
            "billing_cycle": "hourly",
        }

        if ip_addresses:
            params["ip_addresses"] = ip_addresses
        params.update(kwargs)

        if cloud_init:
            params["userdata"] = cloud_init
        data = self.connection.request(
            "/metal/v1/projects/%s/devices" % (ex_project_id),
            data=json.dumps(params),
            method="POST",
        )

        status = data.object.get("status", "OK")

        if status == "ERROR":
            message = data.object.get("message", None)
            error_message = data.object.get("error_message", message)
            raise ValueError("Failed to create node: %s" % (error_message))
        node = self._to_node(data=data.object)

        return node

    def reboot_node(self, node):
        params = {"type": "reboot"}
        res = self.connection.request(
            "/metal/v1/devices/%s/actions" % (node.id), params=params, method="POST"
        )

        return res.status == httplib.OK

    def start_node(self, node):
        params = {"type": "power_on"}
        res = self.connection.request(
            "/metal/v1/devices/%s/actions" % (node.id), params=params, method="POST"
        )

        return res.status == httplib.OK

    def stop_node(self, node):
        params = {"type": "power_off"}
        res = self.connection.request(
            "/metal/v1/devices/%s/actions" % (node.id), params=params, method="POST"
        )

        return res.status == httplib.OK

    def destroy_node(self, node):
        res = self.connection.request("/metal/v1/devices/%s" % (node.id), method="DELETE")

        return res.status == httplib.OK

    def ex_start_node(self, node):
        # NOTE: This method is here for backward compatibility reasons after
        # this method was promoted to be part of the standard compute API in
        # Libcloud v2.7.0

        return self.start_node(node=node)

    def ex_stop_node(self, node):
        # NOTE: This method is here for backward compatibility reasons after
        # this method was promoted to be part of the standard compute API in
        # Libcloud v2.7.0

        return self.stop_node(node=node)

    def ex_reinstall_node(self, node):
        params = {"type": "reinstall"}
        res = self.connection.request(
            "/metal/v1/devices/%s/actions" % (node.id), params=params, method="POST"
        )

        return res.status == httplib.OK

    def ex_rescue_node(self, node):
        params = {"type": "rescue"}
        res = self.connection.request(
            "/metal/v1/devices/%s/actions" % (node.id), params=params, method="POST"
        )

        return res.status == httplib.OK

    def ex_update_node(self, node, **kwargs):
        path = "/metal/v1/devices/%s" % node.id
        res = self.connection.request(path, params=kwargs, method="PUT")

        return res.status == httplib.OK

    def ex_get_node_bandwidth(self, node, from_time, until_time):
        path = "/metal/v1/devices/%s/bandwidth" % node.id
        params = {"from": from_time, "until": until_time}

        return self.connection.request(path, params=params).object

    def ex_list_ip_assignments_for_node(self, node, include=""):
        path = "/metal/v1/devices/%s/ips" % node.id
        params = {"include": include}

        return self.connection.request(path, params=params).object

    def list_key_pairs(self):
        """
        List all the available SSH keys.

        :return: Available SSH keys.
        :rtype: ``list`` of :class:`.KeyPair` objects
        """
        data = self.connection.request("/metal/v1/ssh-keys").object["ssh_keys"]

        return list(map(self._to_key_pairs, data))

    def create_key_pair(self, name, public_key):
        """
        Create a new SSH key.

        :param      name: Key name (required)
        :type       name: ``str``

        :param      public_key: Valid public key string (required)
        :type       public_key: ``str``
        """
        params = {"label": name, "key": public_key}
        data = self.connection.request("/metal/v1/ssh-keys", method="POST", params=params).object

        return self._to_key_pairs(data)

    def delete_key_pair(self, key):
        """
        Delete an existing SSH key.

        :param      key: SSH key (required)
        :type       key: :class:`KeyPair`
        """
        key_id = key.name
        res = self.connection.request("/metal/v1/ssh-keys/%s" % (key_id), method="DELETE")

        return res.status == httplib.NO_CONTENT

    def _to_node(self, data):
        extra = {}
        extra_keys = [
            "created_at",
            "updated_at",
            "userdata",
            "billing_cycle",
            "locked",
            "iqn",
            "locked",
            "project",
            "description",
        ]

        if "state" in data:
            state = self.NODE_STATE_MAP.get(data["state"], NodeState.UNKNOWN)
        else:
            state = NodeState.UNKNOWN

        if "ip_addresses" in data and data["ip_addresses"] is not None:
            ips = self._parse_ips(data["ip_addresses"])
            public_ips = ips["public"]
            private_ips = ips["private"]
        else:
            public_ips = []
            private_ips = []

        if "operating_system" in data and data["operating_system"] is not None:
            image = self._to_image(data["operating_system"])
            extra["operating_system"] = data["operating_system"].get("name")
        else:
            image = None

        if "plan" in data and data["plan"] is not None:
            size = self._to_size(data["plan"])
            extra["plan"] = data["plan"].get("slug")
        else:
            size = None

        if "facility" in data:
            extra["facility"] = data["facility"]

        if "metro" in data and data["metro"] is not None:
            extra["metro"] = data["metro"]

        for key in extra_keys:
            if key in data:
                extra[key] = data[key]

        node = Node(
            id=data["id"],
            name=data["hostname"],
            state=state,
            public_ips=public_ips,
            private_ips=private_ips,
            size=size,
            image=image,
            extra=extra,
            driver=self,
        )

        return node

    def _to_image(self, data):
        extra = {
            "distro": data["distro"],
            "version": data["version"],
            "supported_sizes": data.get("provisionable_on", []),
        }

        return NodeImage(id=data["slug"], name=data["name"], extra=extra, driver=self)

    def _to_location(self, data):
        extra = data

        return NodeLocation(
            id=data["id"], name=data["name"], country=None, driver=self, extra=extra
        )

    def _to_size(self, data):
        try:
            cpus = data["specs"]["cpus"][0].get("count")
        except KeyError:
            cpus = None
        regions = [
            region.get("href").replace("/metal/v1/locations/metros", "")
            for region in data.get("available_in_metros", [])
        ]
        extra = {
            "description": data["description"],
            "line": data["line"],
            "cpus": cpus,
            "regions": regions,
        }
        try:
            factor = 1
            ram_txt = data["specs"]["memory"]["total"]

            if "GB" in ram_txt:
                factor = 1024
                ram_txt = ram_txt.replace("GB", "")
            elif "TB" in ram_txt:
                factor = 1024 * 1024
                ram_txt = ram_txt.replace("TB", "")
            ram = int(ram_txt) * factor
        except Exception:
            ram = None
        disk = None

        if data["specs"].get("drives", ""):
            disk = 0

            for disks in data["specs"]["drives"]:
                disk_size = disks["size"].replace("GB", "")

                if "TB" in disk_size:
                    disk_size = float(disks["size"].replace("TB", "")) * 1000
                disk += disks["count"] * int(disk_size)
        name = "{} - {} RAM".format(data.get("name"), ram)
        price = data["pricing"].get("hour")

        return NodeSize(
            id=data["slug"],
            name=name,
            ram=ram,
            disk=disk,
            bandwidth=0,
            price=price,
            extra=extra,
            driver=self,
        )

    def _to_key_pairs(self, data):
        extra = {
            "label": data["label"],
            "created_at": data["created_at"],
            "updated_at": data["updated_at"],
        }

        return KeyPair(
            name=data["id"],
            fingerprint=data["fingerprint"],
            public_key=data["key"],
            private_key=None,
            driver=self,
            extra=extra,
        )

    def _parse_ips(self, data):
        public_ips = []
        private_ips = []

        for address in data:
            if "address" in address and address["address"] is not None:
                if "public" in address and address["public"] is True:
                    public_ips.append(address["address"])
                else:
                    private_ips.append(address["address"])

        return {"public": public_ips, "private": private_ips}

    def ex_get_bgp_config_for_project(self, ex_project_id):
        path = "/metal/v1/projects/%s/bgp-config" % ex_project_id

        return self.connection.request(path).object

    def ex_get_bgp_config(self, ex_project_id=None):
        if ex_project_id:
            projects = [ex_project_id]
        elif self.project_id:
            projects = [self.project_id]
        else:
            projects = [p.id for p in self.projects]
        retval = []

        for p in projects:
            config = self.ex_get_bgp_config_for_project(p)

            if config:
                retval.append(config)

        return retval

    def ex_get_bgp_session(self, session_uuid):
        path = "/metal/v1/bgp/sessions/%s" % session_uuid

        return self.connection.request(path).object

    def ex_list_bgp_sessions_for_node(self, node):
        path = "/metal/v1/devices/%s/bgp/sessions" % node.id

        return self.connection.request(path).object

    def ex_list_bgp_sessions_for_project(self, ex_project_id):
        path = "/metal/v1/projects/%s/bgp/sessions" % ex_project_id

        return self.connection.request(path).object

    def ex_list_bgp_sessions(self, ex_project_id=None):
        if ex_project_id:
            projects = [ex_project_id]
        elif self.project_id:
            projects = [self.project_id]
        else:
            projects = [p.id for p in self.projects]
        retval = []

        for p in projects:
            retval.extend(self.ex_list_bgp_sessions_for_project(p)["bgp_sessions"])

        return retval

    def ex_create_bgp_session(self, node, address_family="ipv4"):
        path = "/metal/v1/devices/%s/bgp/sessions" % node.id
        params = {"address_family": address_family}
        res = self.connection.request(path, params=params, method="POST")

        return res.object

    def ex_delete_bgp_session(self, session_uuid):
        path = "/metal/v1/bgp/sessions/%s" % session_uuid
        res = self.connection.request(path, method="DELETE")

        return res.status == httplib.OK  # or res.status == httplib.NO_CONTENT

    def ex_list_events_for_node(self, node, include=None, page=1, per_page=10):
        path = "/metal/v1/devices/%s/events" % node.id
        params = {"include": include, "page": page, "per_page": per_page}

        return self.connection.request(path, params=params).object

    def ex_list_events_for_project(self, project, include=None, page=1, per_page=10):
        path = "/metal/v1/projects/%s/events" % project.id
        params = {"include": include, "page": page, "per_page": per_page}

        return self.connection.request(path, params=params).object

    def ex_describe_all_addresses(self, ex_project_id=None, only_associated=False):
        if ex_project_id:
            projects = [ex_project_id]
        elif self.project_id:
            projects = [self.project_id]
        else:
            projects = [p.id for p in self.projects]
        retval = []

        for project in projects:
            retval.extend(self.ex_describe_all_addresses_for_project(project, only_associated))

        return retval

    def ex_describe_all_addresses_for_project(
        self, ex_project_id, include=None, only_associated=False
    ):
        """
        Returns all the reserved IP addresses for this project
        optionally, returns only addresses associated with nodes.

        :param    only_associated: If true, return only the addresses
                                   that are associated with an instance.
        :type     only_associated: ``bool``

        :return:  List of IP addresses.
        :rtype:   ``list`` of :class:`dict`
        """
        path = "/metal/v1/projects/%s/ips" % ex_project_id
        params = {
            "include": include,
        }
        ip_addresses = self.connection.request(path, params=params).object
        result = [
            a
            for a in ip_addresses.get("ip_addresses", [])
            if not only_associated or len(a.get("assignments", [])) > 0
        ]

        return result

    def ex_describe_address(self, ex_address_id, include=None):
        path = "/metal/v1/ips/%s" % ex_address_id
        params = {
            "include": include,
        }
        result = self.connection.request(path, params=params).object

        return result

    def ex_request_address_reservation(
        self,
        ex_project_id,
        location_id=None,
        address_family="global_ipv4",
        quantity=1,
        comments="",
        customdata="",
    ):
        path = "/metal/v1/projects/%s/ips" % ex_project_id
        params = {
            "type": address_family,
            "quantity": quantity,
        }

        if location_id:
            params["metro"] = location_id

        if comments:
            params["comments"] = comments

        if customdata:
            params["customdata"] = customdata
        result = self.connection.request(path, params=params, method="POST").object

        return result

    def ex_associate_address_with_node(self, node, address, manageable=False, customdata=""):
        path = "/metal/v1/devices/%s/ips" % node.id
        params = {
            "address": address,
            "manageable": manageable,
            "customdata": customdata,
        }
        result = self.connection.request(path, params=params, method="POST").object

        return result

    def ex_disassociate_address(self, address_uuid, include=None):
        path = "/metal/v1/ips/%s" % address_uuid
        params = {}

        if include:
            params["include"] = include
        result = self.connection.request(path, params=params, method="DELETE").object

        return result

    def _valid_location(self, metro_code):
        if not metro_code:
            return False
        metros = self.connection.request("/metal/v1/locations/metros").object["metros"]

        for metro in metros:
            if metro["code"] == metro_code:
                return True

        return False


class Project:
    def __init__(self, project):
        self.id = project.get("id")
        self.name = project.get("name")
        self.extra = {}
        self.extra["max_devices"] = project.get("max_devices")
        self.extra["payment_method"] = project.get("payment_method")
        self.extra["created_at"] = project.get("created_at")
        self.extra["credit_amount"] = project.get("credit_amount")
        self.extra["devices"] = project.get("devices")
        self.extra["invitations"] = project.get("invitations")
        self.extra["memberships"] = project.get("memberships")
        self.extra["href"] = project.get("href")
        self.extra["members"] = project.get("members")
        self.extra["ssh_keys"] = project.get("ssh_keys")

    def __repr__(self):
        return ("<Project: id=%s, name=%s>") % (self.id, self.name)



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
VPS.net driver
"""

import base64

from libcloud.utils.py3 import b
from libcloud.common.base import JsonResponse, ConnectionUserAndKey
from libcloud.common.types import InvalidCredsError, MalformedResponseError
from libcloud.compute.base import Node, NodeSize, NodeImage, NodeDriver, NodeLocation
from libcloud.compute.types import NodeState
from libcloud.compute.providers import Provider

try:
    import simplejson as json
except ImportError:
    import json


API_HOST = "api.vps.net"
API_VERSION = "api10json"

RAM_PER_NODE = 256
DISK_PER_NODE = 10
BANDWIDTH_PER_NODE = 250


class VPSNetResponse(JsonResponse):
    def parse_body(self):
        try:
            return super().parse_body()
        except MalformedResponseError:
            return self.body

    def success(self):
        # vps.net wrongly uses 406 for invalid auth creds
        if self.status == 406 or self.status == 403:
            raise InvalidCredsError()
        return True

    def parse_error(self):
        try:
            errors = super().parse_body()["errors"][0]
        except MalformedResponseError:
            return self.body
        else:
            return "\n".join(errors)


class VPSNetConnection(ConnectionUserAndKey):
    """
    Connection class for the VPS.net driver
    """

    host = API_HOST
    responseCls = VPSNetResponse

    allow_insecure = False

    def add_default_headers(self, headers):
        user_b64 = base64.b64encode(b("{}:{}".format(self.user_id, self.key)))
        headers["Authorization"] = "Basic %s" % (user_b64.decode("utf-8"))
        return headers


class VPSNetNodeDriver(NodeDriver):
    """
    VPS.net node driver
    """

    type = Provider.VPSNET
    api_name = "vps_net"
    name = "vps.net"
    website = "http://vps.net/"
    connectionCls = VPSNetConnection

    def _to_node(self, vm):
        if vm["running"]:
            state = NodeState.RUNNING
        else:
            state = NodeState.PENDING

        n = Node(
            id=vm["id"],
            name=vm["label"],
            state=state,
            public_ips=[vm.get("primary_ip_address", None)],
            private_ips=[],
            extra={"slices_count": vm["slices_count"]},
            # Number of nodes consumed by VM
            driver=self.connection.driver,
        )
        return n

    def _to_image(self, image, cloud):
        image = NodeImage(
            id=image["id"],
            name="{}: {}".format(cloud, image["label"]),
            driver=self.connection.driver,
        )

        return image

    def _to_size(self, num):
        size = NodeSize(
            id=num,
            name="%d Node" % (num,),
            ram=RAM_PER_NODE * num,
            disk=DISK_PER_NODE,
            bandwidth=BANDWIDTH_PER_NODE * num,
            price=self._get_price_per_node(num) * num,
            driver=self.connection.driver,
        )
        return size

    def _get_price_per_node(self, num):
        single_node_price = self._get_size_price(size_id="1")
        return num * single_node_price

    def create_node(self, name, image, size, ex_backups_enabled=False, ex_fqdn=None):
        """Create a new VPS.net node

        @inherits: :class:`NodeDriver.create_node`

        :keyword    ex_backups_enabled: Enable automatic backups
        :type       ex_backups_enabled: ``bool``

        :keyword    ex_fqdn:   Fully Qualified domain of the node
        :type       ex_fqdn:   ``str``
        """
        ex_backups_enabled = 1 if ex_backups_enabled else 0
        headers = {"Content-Type": "application/json"}
        request = {
            "virtual_machine": {
                "label": name,
                "fqdn": ex_fqdn or "",
                "system_template_id": image.id,
                "backups_enabled": ex_backups_enabled,
                "slices_required": size.id,
            }
        }

        res = self.connection.request(
            "/virtual_machines.{}".format(API_VERSION),
            data=json.dumps(request),
            headers=headers,
            method="POST",
        )
        node = self._to_node(res.object["virtual_machine"])
        return node

    def reboot_node(self, node):
        res = self.connection.request(
            "/virtual_machines/{}/{}.{}".format(node.id, "reboot", API_VERSION),
            method="POST",
        )
        node = self._to_node(res.object["virtual_machine"])
        return True

    def list_sizes(self, location=None):
        res = self.connection.request("/nodes.{}".format(API_VERSION))
        available_nodes = len([size for size in res.object if size["slice"]["virtual_machine_id"]])
        sizes = [self._to_size(i) for i in range(1, available_nodes + 1)]
        return sizes

    def destroy_node(self, node):
        res = self.connection.request(
            "/virtual_machines/{}.{}".format(node.id, API_VERSION), method="DELETE"
        )
        return res.status == 200

    def list_nodes(self):
        res = self.connection.request("/virtual_machines.{}".format(API_VERSION))
        return [self._to_node(i["virtual_machine"]) for i in res.object]

    def list_images(self, location=None):
        res = self.connection.request("/available_clouds.{}".format(API_VERSION))

        images = []
        for cloud in res.object:
            label = cloud["cloud"]["label"]
            templates = cloud["cloud"]["system_templates"]
            images.extend([self._to_image(image, label) for image in templates])

        return images

    def list_locations(self):
        return [NodeLocation(0, "VPS.net Western US", "US", self)]



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from libcloud.common.types import LibcloudError
from libcloud.storage.providers import Provider
from libcloud.storage.drivers.s3 import BaseS3Connection, BaseS3StorageDriver

__all__ = ["AuroraObjectsStorageDriver"]

AURORA_OBJECTS_EU_HOST = "o.auroraobjects.eu"

NO_CDN_SUPPORT_ERROR = "CDN is not supported by AuroraObjects"


class BaseAuroraObjectsConnection(BaseS3Connection):
    host = AURORA_OBJECTS_EU_HOST


class BaseAuroraObjectsStorageDriver(BaseS3StorageDriver):
    type = Provider.AURORAOBJECTS
    name = "PCextreme AuroraObjects"
    website = "https://www.pcextreme.com/aurora/objects"


class AuroraObjectsStorageDriver(BaseAuroraObjectsStorageDriver):
    connectionCls = BaseAuroraObjectsConnection

    def enable_container_cdn(self, *argv):
        raise LibcloudError(NO_CDN_SUPPORT_ERROR, driver=self)

    def enable_object_cdn(self, *argv):
        raise LibcloudError(NO_CDN_SUPPORT_ERROR, driver=self)

    def get_container_cdn_url(self, *argv):
        raise LibcloudError(NO_CDN_SUPPORT_ERROR, driver=self)

    def get_object_cdn_url(self, *argv):
        raise LibcloudError(NO_CDN_SUPPORT_ERROR, driver=self)



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Base types used by other parts of libcloud
"""

from libcloud.common.types import (
    Type,
    LibcloudError,
    InvalidCredsError,
    InvalidCredsException,
    MalformedResponseError,
)

__all__ = [
    "Provider",
    "NodeState",
    "DeploymentError",
    "DeploymentException",
    # @@TR: should the unused imports below be exported?
    "LibcloudError",
    "MalformedResponseError",
    "InvalidCredsError",
    "InvalidCredsException",
    "OLD_CONSTANT_TO_NEW_MAPPING",
]


class Provider(Type):
    """
    Defines for each of the supported providers

    Non-Dummy drivers are sorted in alphabetical order. Please preserve this
    ordering when adding new drivers.

    :cvar DUMMY: Example provider
    :cvar ABIQUO: Abiquo driver
    :cvar ALIYUN_ECS: Aliyun ECS driver.
    :cvar AURORACOMPUTE: Aurora Compute driver.
    :cvar AZURE: Azure (classic) driver.
    :cvar AZURE_ARM: Azure Resource Manager (modern) driver.
    :cvar CLOUDSIGMA: CloudSigma
    :cvar CLOUDSCALE: cloudscale.ch
    :cvar CLOUDSTACK: CloudStack
    :cvar DIMENSIONDATA: Dimension Data Cloud
    :cvar EC2: Amazon AWS.
    :cvar EXOSCALE: Exoscale driver.
    :cvar GCE: Google Compute Engine
    :cvar GRIDSCALE: gridscale
    :cvar IBM: IBM Developer Cloud
    :cvar IKOULA: Ikoula driver.
    :cvar KAMATERA: Kamatera driver
    :cvar KTUCLOUD: kt ucloud driver
    :cvar KUBEVIRT: kubevirt driver
    :cvar LIBVIRT: Libvirt driver
    :cvar LINODE: Linode.com
    :cvar NIMBUS: Nimbus
    :cvar NINEFOLD: Ninefold
    :cvar NTTC-CIS: NTT Communications CIS
    :cvar OPENNEBULA: OpenNebula.org
    :cvar OPSOURCE: Opsource Cloud
    :cvar OUTSCALE_INC: Outscale INC driver.
    :cvar OUTSCALE_SAS: Outscale SAS driver.
    :cvar OUTSCALE_SDK: Outscale SDK driver.
    :cvar RACKSPACE: Rackspace next-gen OpenStack based Cloud Servers
    :cvar RACKSPACE_FIRST_GEN: Rackspace First Gen Cloud Servers
    :cvar RIMUHOSTING: RimuHosting.com
    :cvar TERREMARK: Terremark
    :cvar UPCLOUD: UpCloud
    :cvar VCL: VCL driver
    :cvar VCLOUD: vmware vCloud
    :cvar VPSNET: VPS.net
    :cvar VSphere: VSphere driver.
    :cvar VULTR: vultr driver.
    """

    AZURE = "azure"
    AZURE_ARM = "azure_arm"
    DUMMY = "dummy"
    ABIQUO = "abiquo"
    ALIYUN_ECS = "aliyun_ecs"
    AURORACOMPUTE = "aurora_compute"
    BRIGHTBOX = "brightbox"
    CISCOCCS = "ciscoccs"
    CLOUDFRAMES = "cloudframes"
    CLOUDSIGMA = "cloudsigma"
    CLOUDSCALE = "cloudscale"
    CLOUDSTACK = "cloudstack"
    DIGITAL_OCEAN = "digitalocean"
    DIMENSIONDATA = "dimensiondata"
    EC2 = "ec2"
    EQUINIXMETAL = "equinixmetal"
    EUCALYPTUS = "eucalyptus"
    EXOSCALE = "exoscale"
    GANDI = "gandi"
    GCE = "gce"
    GIG_G8 = "gig_g8"
    GRIDSCALE = "gridscale"
    IBM = "ibm"
    IKOULA = "ikoula"
    INTERNETSOLUTIONS = "internetsolutions"
    KAMATERA = "kamatera"
    KTUCLOUD = "ktucloud"
    KUBEVIRT = "kubevirt"
    LIBVIRT = "libvirt"
    LINODE = "linode"
    MAXIHOST = "maxihost"
    NIMBUS = "nimbus"
    NINEFOLD = "ninefold"
    NTTA = "ntta"
    NTTCIS = "nttcis"
    OPENNEBULA = "opennebula"
    OPENSTACK = "openstack"
    OPSOURCE = "opsource"
    OUTSCALE_INC = "outscale_inc"
    OUTSCALE_SAS = "outscale_sas"
    OUTSCALE = "outscale"
    OVH = "ovh"
    RACKSPACE = "rackspace"
    RACKSPACE_FIRST_GEN = "rackspace_first_gen"
    RIMUHOSTING = "rimuhosting"
    RUNABOVE = "runabove"
    SCALEWAY = "scaleway"
    TERREMARK = "terremark"
    UPCLOUD = "upcloud"
    VCL = "vcl"
    VCLOUD = "vcloud"
    VPSNET = "vpsnet"
    VSPHERE = "vsphere"
    VULTR = "vultr"

    # OpenStack based providers
    HPCLOUD = "hpcloud"
    ONAPP = "onapp"

    # Deprecated constants which aren't supported anymore
    RACKSPACE_UK = "rackspace_uk"
    RACKSPACE_NOVA_BETA = "rackspace_nova_beta"
    RACKSPACE_NOVA_DFW = "rackspace_nova_dfw"
    RACKSPACE_NOVA_LON = "rackspace_nova_lon"
    RACKSPACE_NOVA_ORD = "rackspace_nova_ord"

    EC2_US_EAST = "ec2_us_east"
    EC2_US_EAST_OHIO = "ec2_us_east_ohio"
    EC2_EU = "ec2_eu_west"  # deprecated name
    EC2_EU_WEST = "ec2_eu_west"
    EC2_EU_WEST2 = "ec2_eu_west_london"
    EC2_US_WEST = "ec2_us_west"
    EC2_AP_SOUTHEAST = "ec2_ap_southeast"
    EC2_AP_NORTHEAST = "ec2_ap_northeast"
    EC2_AP_NORTHEAST1 = "ec2_ap_northeast_1"
    EC2_AP_NORTHEAST2 = "ec2_ap_northeast_2"
    EC2_US_WEST_OREGON = "ec2_us_west_oregon"
    EC2_SA_EAST = "ec2_sa_east"
    EC2_AP_SOUTHEAST2 = "ec2_ap_southeast_2"
    EC2_CA_CENTRAL1 = "ec2_ca_central_1"

    CLOUDSIGMA_US = "cloudsigma_us"

    # Removed
    # SLICEHOST = 'slicehost'


DEPRECATED_RACKSPACE_PROVIDERS = [
    Provider.RACKSPACE_UK,
    Provider.RACKSPACE_NOVA_BETA,
    Provider.RACKSPACE_NOVA_DFW,
    Provider.RACKSPACE_NOVA_LON,
    Provider.RACKSPACE_NOVA_ORD,
]
OLD_CONSTANT_TO_NEW_MAPPING = {
    # Rackspace
    Provider.RACKSPACE_UK: Provider.RACKSPACE_FIRST_GEN,
    Provider.RACKSPACE_NOVA_BETA: Provider.RACKSPACE,
    Provider.RACKSPACE_NOVA_DFW: Provider.RACKSPACE,
    Provider.RACKSPACE_NOVA_LON: Provider.RACKSPACE,
    Provider.RACKSPACE_NOVA_ORD: Provider.RACKSPACE,
    # AWS
    Provider.EC2_US_EAST: Provider.EC2,
    Provider.EC2_US_EAST_OHIO: Provider.EC2,
    Provider.EC2_EU: Provider.EC2,
    Provider.EC2_EU_WEST: Provider.EC2,
    Provider.EC2_EU_WEST2: Provider.EC2,
    Provider.EC2_US_WEST: Provider.EC2,
    Provider.EC2_AP_SOUTHEAST: Provider.EC2,
    Provider.EC2_AP_SOUTHEAST2: Provider.EC2,
    Provider.EC2_AP_NORTHEAST: Provider.EC2,
    Provider.EC2_AP_NORTHEAST1: Provider.EC2,
    Provider.EC2_AP_NORTHEAST2: Provider.EC2,
    Provider.EC2_US_WEST_OREGON: Provider.EC2,
    Provider.EC2_SA_EAST: Provider.EC2,
    Provider.EC2_CA_CENTRAL1: Provider.EC2,
}


class NodeState(Type):
    """
    Standard states for a node

    :cvar RUNNING: Node is running.
    :cvar STARTING: Node is starting up.
    :cvar REBOOTING: Node is rebooting.
    :cvar TERMINATED: Node is terminated. This node can't be started later on.
    :cvar STOPPING: Node is currently trying to stop.
    :cvar STOPPED: Node is stopped. This node can be started later on.
    :cvar PENDING: Node is pending.
    :cvar SUSPENDED: Node is suspended.
    :cvar ERROR: Node is an error state. Usually no operations can be performed
                 on the node once it ends up in the error state.
    :cvar PAUSED: Node is paused.
    :cvar RECONFIGURING: Node is being reconfigured.
    :cvar UNKNOWN: Node state is unknown.
    """

    RUNNING = "running"
    STARTING = "starting"
    REBOOTING = "rebooting"
    TERMINATED = "terminated"
    PENDING = "pending"
    UNKNOWN = "unknown"
    STOPPING = "stopping"
    STOPPED = "stopped"
    SUSPENDED = "suspended"
    ERROR = "error"
    PAUSED = "paused"
    RECONFIGURING = "reconfiguring"
    MIGRATING = "migrating"
    NORMAL = "normal"
    UPDATING = "updating"


class StorageVolumeState(Type):
    """
    Standard states of a StorageVolume
    """

    AVAILABLE = "available"
    ERROR = "error"
    INUSE = "inuse"
    CREATING = "creating"
    DELETING = "deleting"
    DELETED = "deleted"
    BACKUP = "backup"
    ATTACHING = "attaching"
    UNKNOWN = "unknown"
    MIGRATING = "migrating"
    UPDATING = "updating"


class VolumeSnapshotState(Type):
    """
    Standard states of VolumeSnapshots
    """

    AVAILABLE = "available"
    ERROR = "error"
    CREATING = "creating"
    DELETING = "deleting"
    RESTORING = "restoring"
    UNKNOWN = "unknown"
    UPDATING = "updating"


class NodeImageMemberState(Type):
    """
    Standard states of VolumeSnapshots
    """

    ACCEPTED = "accepted"
    PENDING = "pending"
    REJECTED = "rejected"


class Architecture:
    """
    Image and size architectures.

    :cvar I386: i386 (32 bt)
    :cvar X86_64: x86_64 (64 bit)
    """

    I386 = 0
    X86_X64 = 1


class DeploymentError(LibcloudError):
    """
    Exception used when a Deployment Task failed.

    :ivar node: :class:`Node` on which this exception happened, you might want
                to call :func:`Node.destroy`
    """

    def __init__(self, node, original_exception=None, driver=None):
        self.node = node
        self.value = original_exception
        self.original_error = original_exception
        self.driver = driver

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return "<DeploymentError: node={}, error={}, driver={}>".format(
            self.node.id,
            str(self.value),
            str(self.driver),
        )


class KeyPairError(LibcloudError):
    error_type = "KeyPairError"

    def __init__(self, name, driver):
        self.name = name
        self.value = "Key pair with name %s does not exist" % (name)
        super().__init__(value=self.value, driver=driver)

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return "<{} name={}, value={}, driver={}>".format(
            self.error_type,
            self.name,
            self.value,
            self.driver.name,
        )


class KeyPairDoesNotExistError(KeyPairError):
    error_type = "KeyPairDoesNotExistError"


"""Deprecated alias of :class:`DeploymentException`"""
DeploymentException = DeploymentError



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Gandi driver base classes
"""

import time
import hashlib

from libcloud.utils.py3 import b
from libcloud.common.base import ConnectionKey
from libcloud.common.xmlrpc import XMLRPCResponse, XMLRPCConnection

# Global constants

DEFAULT_TIMEOUT = 600  # operation pooling max seconds
DEFAULT_INTERVAL = 20  # seconds between 2 operation.info


class GandiException(Exception):
    """
    Exception class for Gandi driver
    """

    def __str__(self):
        # pylint: disable=unsubscriptable-object

        return "({}) {}".format(self.args[0], self.args[1])

    def __repr__(self):
        # pylint: disable=unsubscriptable-object

        return '<GandiException code {} "{}">'.format(self.args[0], self.args[1])


class GandiResponse(XMLRPCResponse):
    """
    A Base Gandi Response class to derive from.
    """


class GandiConnection(XMLRPCConnection, ConnectionKey):
    """
    Connection class for the Gandi driver
    """

    responseCls = GandiResponse
    host = "rpc.gandi.net"
    endpoint = "/xmlrpc/"

    def __init__(
        self,
        key,
        secure=True,
        timeout=None,
        retry_delay=None,
        backoff=None,
        proxy_url=None,
    ):
        # Note: Method resolution order in this case is
        # XMLRPCConnection -> Connection and Connection doesn't take key as the
        # first argument so we specify a keyword argument instead.
        # Previously it was GandiConnection -> ConnectionKey so it worked fine.
        super().__init__(
            key=key,
            secure=secure,
            timeout=timeout,
            retry_delay=retry_delay,
            backoff=backoff,
            proxy_url=proxy_url,
        )
        self.driver = BaseGandiDriver

    def request(self, method, *args):
        args = (self.key,) + args

        return super().request(method, *args)


class BaseGandiDriver:
    """
    Gandi base driver

    """

    connectionCls = GandiConnection
    name = "Gandi"

    # Specific methods for gandi
    def _wait_operation(self, id, timeout=DEFAULT_TIMEOUT, check_interval=DEFAULT_INTERVAL):
        """Wait for an operation to succeed"""

        for i in range(0, timeout, check_interval):
            try:
                # pylint: disable=no-member
                op = self.connection.request("operation.info", int(id)).object

                if op["step"] == "DONE":
                    return True

                if op["step"] in ["ERROR", "CANCEL"]:
                    return False
            except (KeyError, IndexError):
                pass
            except Exception as e:
                raise GandiException(1002, e)

            time.sleep(check_interval)

        return False


class BaseObject:
    """Base class for objects not conventional"""

    uuid_prefix = ""

    def __init__(self, id, state, driver):
        self.id = str(id) if id else None
        self.state = state
        self.driver = driver
        self.uuid = self.get_uuid()

    def get_uuid(self):
        """Unique hash for this object

        :return: ``str``

        The hash is a function of an SHA1 hash of prefix, the object's ID and
        its driver which means that it should be unique between all
        interfaces.
        TODO : to review
        >>> from libcloud.compute.drivers.dummy import DummyNodeDriver
        >>> driver = DummyNodeDriver(0)
        >>> vif = driver.create_interface()
        >>> vif.get_uuid()
        'd3748461511d8b9b0e0bfa0d4d3383a619a2bb9f'

        Note, for example, that this example will always produce the
        same UUID!
        """
        hashstring = "{}:{}:{}".format(self.uuid_prefix, self.id, self.driver.type)

        return hashlib.sha1(b(hashstring)).hexdigest()  # nosec


class IPAddress(BaseObject):
    """
    Provide a common interface for ip addresses
    """

    uuid_prefix = "inet:"

    def __init__(self, id, state, inet, driver, version=4, extra=None):
        super().__init__(id, state, driver)
        self.inet = inet
        self.version = version
        self.extra = extra or {}

    def __repr__(self):
        return ("<IPAddress: id=%s, address=%s, state=%s, driver=%s ...>") % (
            self.id,
            self.inet,
            self.state,
            self.driver.name,
        )


class NetworkInterface(BaseObject):
    """
    Provide a common interface for network interfaces
    """

    uuid_prefix = "if:"

    def __init__(self, id, state, mac_address, driver, ips=None, node_id=None, extra=None):
        super().__init__(id, state, driver)
        self.mac = mac_address
        self.ips = ips or {}
        self.node_id = node_id
        self.extra = extra or {}

    def __repr__(self):
        return ("<Interface: id=%s, mac=%s, state=%s, driver=%s ...>") % (
            self.id,
            self.mac,
            self.state,
            self.driver.name,
        )


class Disk(BaseObject):
    """
    Gandi disk component
    """

    def __init__(self, id, state, name, driver, size, extra=None):
        super().__init__(id, state, driver)
        self.name = name
        self.size = size
        self.extra = extra or {}

    def __repr__(self):
        return ("<Disk: id=%s, name=%s, state=%s, size=%s, driver=%s ...>") % (
            self.id,
            self.name,
            self.state,
            self.size,
            self.driver.name,
        )



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from libcloud.utils.misc import reverse_dict
from libcloud.common.cloudstack import CloudStackDriverMixIn
from libcloud.loadbalancer.base import DEFAULT_ALGORITHM, Driver, Member, Algorithm, LoadBalancer
from libcloud.loadbalancer.types import State, Provider


class CloudStackLBDriver(CloudStackDriverMixIn, Driver):
    """Driver for CloudStack load balancers."""

    api_name = "cloudstack_lb"
    name = "CloudStack"
    website = "http://cloudstack.org/"
    type = Provider.CLOUDSTACK

    _VALUE_TO_ALGORITHM_MAP = {
        "roundrobin": Algorithm.ROUND_ROBIN,
        "leastconn": Algorithm.LEAST_CONNECTIONS,
    }
    _ALGORITHM_TO_VALUE_MAP = reverse_dict(_VALUE_TO_ALGORITHM_MAP)

    LB_STATE_MAP = {
        "Active": State.RUNNING,
    }

    def __init__(
        self,
        key,
        secret=None,
        secure=True,
        host=None,
        path=None,
        port=None,
        *args,
        **kwargs,
    ):
        """
        @inherits: :class:`Driver.__init__`
        """
        host = host if host else self.host
        path = path if path else self.path

        if path is not None:
            self.path = path

        if host is not None:
            self.host = host

        if (self.type == Provider.CLOUDSTACK) and (not host or not path):
            raise Exception(
                "When instantiating CloudStack driver directly "
                + "you also need to provide host and path argument"
            )

        super().__init__(key=key, secret=secret, secure=secure, host=host, port=port)

    def list_protocols(self):
        """
        We don't actually have any protocol awareness beyond TCP.

        :rtype: ``list`` of ``str``
        """
        return ["tcp"]

    def list_balancers(self):
        balancers = self._sync_request(command="listLoadBalancerRules", method="GET")
        balancers = balancers.get("loadbalancerrule", [])
        return [self._to_balancer(balancer) for balancer in balancers]

    def get_balancer(self, balancer_id):
        balancer = self._sync_request(
            command="listLoadBalancerRules", params={"id": balancer_id}, method="GET"
        )
        balancer = balancer.get("loadbalancerrule", [])
        if not balancer:
            raise Exception("no such load balancer: " + str(balancer_id))
        return self._to_balancer(balancer[0])

    def create_balancer(
        self,
        name,
        members,
        protocol="http",
        port=80,
        algorithm=DEFAULT_ALGORITHM,
        location=None,
        private_port=None,
        network_id=None,
        vpc_id=None,
    ):
        """
        @inherits: :class:`Driver.create_balancer`

        :param location: Location
        :type  location: :class:`NodeLocation`

        :param private_port: Private port
        :type  private_port: ``int``

        :param network_id: The guest network this rule will be created for.
        :type  network_id: ``str``
        """

        args = {}
        ip_args = {}

        if location is None:
            locations = self._sync_request(command="listZones", method="GET")
            location = locations["zone"][0]["id"]
        else:
            location = location.id
        if private_port is None:
            private_port = port

        if network_id is not None:
            args["networkid"] = network_id
            ip_args["networkid"] = network_id

        if vpc_id is not None:
            ip_args["vpcid"] = vpc_id

        ip_args.update({"zoneid": location, "networkid": network_id, "vpc_id": vpc_id})

        result = self._async_request(command="associateIpAddress", params=ip_args, method="GET")
        public_ip = result["ipaddress"]

        args.update(
            {
                "algorithm": self._ALGORITHM_TO_VALUE_MAP[algorithm],
                "name": name,
                "privateport": private_port,
                "publicport": port,
                "publicipid": public_ip["id"],
            }
        )

        result = self._sync_request(command="createLoadBalancerRule", params=args, method="GET")

        listbalancers = self._sync_request(
            command="listLoadBalancerRules", params=args, method="GET"
        )

        listbalancers = [
            rule for rule in listbalancers["loadbalancerrule"] if rule["id"] == result["id"]
        ]
        if len(listbalancers) != 1:
            return None

        balancer = self._to_balancer(listbalancers[0])

        for member in members:
            balancer.attach_member(member)

        return balancer

    def destroy_balancer(self, balancer):
        self._async_request(
            command="deleteLoadBalancerRule", params={"id": balancer.id}, method="GET"
        )
        self._async_request(
            command="disassociateIpAddress",
            params={"id": balancer.ex_public_ip_id},
            method="GET",
        )

    def balancer_attach_member(self, balancer, member):
        member.port = balancer.ex_private_port
        self._async_request(
            command="assignToLoadBalancerRule",
            params={"id": balancer.id, "virtualmachineids": member.id},
            method="GET",
        )
        return True

    def balancer_detach_member(self, balancer, member):
        self._async_request(
            command="removeFromLoadBalancerRule",
            params={"id": balancer.id, "virtualmachineids": member.id},
            method="GET",
        )
        return True

    def balancer_list_members(self, balancer):
        members = self._sync_request(
            command="listLoadBalancerRuleInstances",
            params={"id": balancer.id},
            method="GET",
        )
        members = members["loadbalancerruleinstance"]
        return [self._to_member(m, balancer.ex_private_port, balancer) for m in members]

    def _to_balancer(self, obj):
        balancer = LoadBalancer(
            id=obj["id"],
            name=obj["name"],
            state=self.LB_STATE_MAP.get(obj["state"], State.UNKNOWN),
            ip=obj["publicip"],
            port=obj["publicport"],
            driver=self.connection.driver,
        )
        balancer.ex_private_port = obj["privateport"]
        balancer.ex_public_ip_id = obj["publicipid"]
        return balancer

    def _to_member(self, obj, port, balancer):
        return Member(id=obj["id"], ip=obj["nic"][0]["ipaddress"], port=port, balancer=balancer)



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
libcloud provides a unified interface to the cloud computing resources.

:var __version__: Current version of libcloud
"""

import os
import atexit
import codecs

from libcloud.base import DriverType  # NOQA
from libcloud.base import DriverTypeFactoryMap  # NOQA
from libcloud.base import get_driver  # NOQA

try:
    # TODO: This import is slow and adds overhead in situations when no
    # requests are made but it's necessary for detecting bad version of
    # requests
    import requests  # NOQA

    have_requests = True
except ImportError:
    have_requests = False

__all__ = ["__version__", "enable_debug"]

__version__ = "3.9.1"


def enable_debug(fo):
    """
    Enable library wide debugging to a file-like object.

    :param fo: Where to append debugging information
    :type fo: File like object, only write operations are used.
    """
    from libcloud.common.base import Connection
    from libcloud.utils.loggingconnection import LoggingConnection

    LoggingConnection.log = fo
    Connection.conn_class = LoggingConnection

    # Ensure the file handle is closed on exit
    def close_file(fd):
        try:
            fd.close()
        except Exception:
            pass

    atexit.register(close_file, fo)


def reset_debug():
    """
    Reset debugging functionality (if set).

    NOTE: This function is only meant to be used in the tests.
    """
    from libcloud.common.base import Connection, LibcloudConnection
    from libcloud.utils.loggingconnection import LoggingConnection

    LoggingConnection.log = None
    Connection.conn_class = LibcloudConnection


def _init_once():
    """
    Utility function that is ran once on Library import.

    This checks for the LIBCLOUD_DEBUG environment variable, which if it exists
    is where we will log debug information about the provider transports.

    This also checks for known environment/dependency incompatibilities.
    """
    path = os.getenv("LIBCLOUD_DEBUG")

    if path:
        mode = "a"

        # Special case for /dev/stderr and /dev/stdout on Python 3.
        from libcloud.utils.py3 import PY3

        # Opening those files in append mode will throw "illegal seek"
        # exception there.
        # Late import to avoid setup.py related side affects
        if path in ["/dev/stderr", "/dev/stdout"] and PY3:
            mode = "w"

        fo = codecs.open(path, mode, encoding="utf8")
        enable_debug(fo)

        # NOTE: We use lazy import to avoid unnecessary import time overhead
        try:
            import paramiko  # NOQA

            have_paramiko = True
        except ImportError:
            have_paramiko = False

        if have_paramiko and hasattr(paramiko.util, "log_to_file"):
            import logging

            # paramiko always tries to open file path in append mode which
            # won't work with /dev/{stdout, stderr} so we just ignore those
            # errors
            try:
                paramiko.util.log_to_file(filename=path, level=logging.DEBUG)
            except OSError as e:
                if "illegal seek" not in str(e).lower():
                    raise e

    # check for broken `yum install python-requests`
    if have_requests and requests.__version__ == "2.6.0":
        chardet_version = requests.packages.chardet.__version__
        required_chardet_version = "2.3.0"
        assert chardet_version == required_chardet_version, (
            "Known bad version of requests detected! This can happen when "
            "requests was installed from a source other than PyPI, e.g. via "
            "a package manager such as yum. Please either install requests "
            "from PyPI or run `pip install chardet==%s` to resolve this "
            "issue." % required_chardet_version
        )


_init_once()



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from libcloud.utils.py3 import ET
from libcloud.utils.xml import findall, findtext, fixxpath
from libcloud.backup.base import BackupDriver, BackupTarget, BackupTargetJob
from libcloud.backup.types import Provider, BackupTargetType
from libcloud.common.dimensiondata import (
    BACKUP_NS,
    TYPES_URN,
    GENERAL_NS,
    API_ENDPOINTS,
    DEFAULT_REGION,
    DimensionDataConnection,
    DimensionDataBackupClient,
    DimensionDataBackupDetails,
    DimensionDataBackupClientType,
    DimensionDataBackupClientAlert,
    DimensionDataBackupStoragePolicy,
    DimensionDataBackupSchedulePolicy,
    dd_object_to_id,
)

# pylint: disable=no-member

DEFAULT_BACKUP_PLAN = "Advanced"


class DimensionDataBackupDriver(BackupDriver):
    """
    DimensionData backup driver.
    """

    selected_region = None
    connectionCls = DimensionDataConnection
    name = "Dimension Data Backup"
    website = "https://cloud.dimensiondata.com/"
    type = Provider.DIMENSIONDATA
    api_version = 1.0

    network_domain_id = None

    def __init__(
        self,
        key,
        secret=None,
        secure=True,
        host=None,
        port=None,
        api_version=None,
        region=DEFAULT_REGION,
        **kwargs,
    ):
        if region not in API_ENDPOINTS and host is None:
            raise ValueError("Invalid region: %s, no host specified" % (region))
        if region is not None:
            self.selected_region = API_ENDPOINTS[region]

        super().__init__(
            key=key,
            secret=secret,
            secure=secure,
            host=host,
            port=port,
            api_version=api_version,
            region=region,
            **kwargs,
        )

    def _ex_connection_class_kwargs(self):
        """
        Add the region to the kwargs before the connection is instantiated
        """

        kwargs = super()._ex_connection_class_kwargs()
        kwargs["region"] = self.selected_region
        return kwargs

    def get_supported_target_types(self):
        """
        Get a list of backup target types this driver supports

        :return: ``list`` of :class:``BackupTargetType``
        """
        return [BackupTargetType.VIRTUAL]

    def list_targets(self):
        """
        List all backuptargets

        :rtype: ``list`` of :class:`BackupTarget`
        """
        targets = self._to_targets(self.connection.request_with_orgId_api_2("server/server").object)
        return targets

    def create_target(self, name, address, type=BackupTargetType.VIRTUAL, extra=None):
        """
        Creates a new backup target

        :param name: Name of the target (not used)
        :type name: ``str``

        :param address: The ID of the node in Dimension Data Cloud
        :type address: ``str``

        :param type: Backup target type, only Virtual supported
        :type type: :class:`BackupTargetType`

        :param extra: (optional) Extra attributes (driver specific).
        :type extra: ``dict``

        :rtype: Instance of :class:`BackupTarget`
        """
        if extra is not None:
            service_plan = extra.get("servicePlan", DEFAULT_BACKUP_PLAN)
        else:
            service_plan = DEFAULT_BACKUP_PLAN
            extra = {"servicePlan": service_plan}

        create_node = ET.Element("NewBackup", {"xmlns": BACKUP_NS})
        create_node.set("servicePlan", service_plan)

        response = self.connection.request_with_orgId_api_1(
            "server/%s/backup" % (address), method="POST", data=ET.tostring(create_node)
        ).object

        asset_id = None
        for info in findall(response, "additionalInformation", GENERAL_NS):
            if info.get("name") == "assetId":
                asset_id = findtext(info, "value", GENERAL_NS)

        return BackupTarget(
            id=asset_id, name=name, address=address, type=type, extra=extra, driver=self
        )

    def create_target_from_node(self, node, type=BackupTargetType.VIRTUAL, extra=None):
        """
        Creates a new backup target from an existing node

        :param node: The Node to backup
        :type  node: ``Node``

        :param type: Backup target type (Physical, Virtual, ...).
        :type type: :class:`BackupTargetType`

        :param extra: (optional) Extra attributes (driver specific).
        :type extra: ``dict``

        :rtype: Instance of :class:`BackupTarget`
        """
        return self.create_target(
            name=node.name, address=node.id, type=BackupTargetType.VIRTUAL, extra=extra
        )

    def create_target_from_container(self, container, type=BackupTargetType.OBJECT, extra=None):
        """
        Creates a new backup target from an existing storage container

        :param node: The Container to backup
        :type  node: ``Container``

        :param type: Backup target type (Physical, Virtual, ...).
        :type type: :class:`BackupTargetType`

        :param extra: (optional) Extra attributes (driver specific).
        :type extra: ``dict``

        :rtype: Instance of :class:`BackupTarget`
        """
        return NotImplementedError("create_target_from_container not supported for this driver")

    def update_target(self, target, name=None, address=None, extra=None):
        """
        Update the properties of a backup target, only changing the serviceplan
        is supported.

        :param target: Backup target to update
        :type  target: Instance of :class:`BackupTarget` or ``str``

        :param name: Name of the target
        :type name: ``str``

        :param address: Hostname, FQDN, IP, file path etc.
        :type address: ``str``

        :param extra: (optional) Extra attributes (driver specific).
        :type extra: ``dict``

        :rtype: Instance of :class:`BackupTarget`
        """
        if extra is not None:
            service_plan = extra.get("servicePlan", DEFAULT_BACKUP_PLAN)
        else:
            service_plan = DEFAULT_BACKUP_PLAN
        request = ET.Element("ModifyBackup", {"xmlns": BACKUP_NS})
        request.set("servicePlan", service_plan)
        server_id = self._target_to_target_address(target)
        self.connection.request_with_orgId_api_1(
            "server/%s/backup/modify" % (server_id),
            method="POST",
            data=ET.tostring(request),
        ).object
        if isinstance(target, BackupTarget):
            target.extra = extra
        else:
            target = self.ex_get_target_by_id(server_id)
        return target

    def delete_target(self, target):
        """
        Delete a backup target

        :param target: Backup target to delete
        :type  target: Instance of :class:`BackupTarget` or ``str``

        :rtype: ``bool``
        """
        server_id = self._target_to_target_address(target)
        response = self.connection.request_with_orgId_api_1(
            "server/%s/backup?disable" % (server_id), method="GET"
        ).object
        response_code = findtext(response, "result", GENERAL_NS)
        return response_code in ["IN_PROGRESS", "SUCCESS"]

    def list_recovery_points(self, target, start_date=None, end_date=None):
        """
        List the recovery points available for a target

        :param target: Backup target to delete
        :type  target: Instance of :class:`BackupTarget`

        :param start_date: The start date to show jobs between (optional)
        :type  start_date: :class:`datetime.datetime`

        :param end_date: The end date to show jobs between (optional)
        :type  end_date: :class:`datetime.datetime``

        :rtype: ``list`` of :class:`BackupTargetRecoveryPoint`
        """
        raise NotImplementedError("list_recovery_points not implemented for this driver")

    def recover_target(self, target, recovery_point, path=None):
        """
        Recover a backup target to a recovery point

        :param target: Backup target to delete
        :type  target: Instance of :class:`BackupTarget`

        :param recovery_point: Backup target with the backup data
        :type  recovery_point: Instance of :class:`BackupTarget`

        :param path: The part of the recovery point to recover (optional)
        :type  path: ``str``

        :rtype: Instance of :class:`BackupTargetJob`
        """
        raise NotImplementedError("recover_target not implemented for this driver")

    def recover_target_out_of_place(self, target, recovery_point, recovery_target, path=None):
        """
        Recover a backup target to a recovery point out-of-place

        :param target: Backup target with the backup data
        :type  target: Instance of :class:`BackupTarget`

        :param recovery_point: Backup target with the backup data
        :type  recovery_point: Instance of :class:`BackupTarget`

        :param recovery_target: Backup target with to recover the data to
        :type  recovery_target: Instance of :class:`BackupTarget`

        :param path: The part of the recovery point to recover (optional)
        :type  path: ``str``

        :rtype: Instance of :class:`BackupTargetJob`
        """
        raise NotImplementedError("recover_target_out_of_place not implemented for this driver")

    def get_target_job(self, target, id):
        """
        Get a specific backup job by ID

        :param target: Backup target with the backup data
        :type  target: Instance of :class:`BackupTarget`

        :param id: Backup target with the backup data
        :type  id: Instance of :class:`BackupTarget`

        :rtype: :class:`BackupTargetJob`
        """
        jobs = self.list_target_jobs(target)
        return list(filter(lambda x: x.id == id, jobs))[0]

    def list_target_jobs(self, target):
        """
        List the backup jobs on a target

        :param target: Backup target with the backup data
        :type  target: Instance of :class:`BackupTarget`

        :rtype: ``list`` of :class:`BackupTargetJob`
        """
        raise NotImplementedError("list_target_jobs not implemented for this driver")

    def create_target_job(self, target, extra=None):
        """
        Create a new backup job on a target

        :param target: Backup target with the backup data
        :type  target: Instance of :class:`BackupTarget`

        :param extra: (optional) Extra attributes (driver specific).
        :type extra: ``dict``

        :rtype: Instance of :class:`BackupTargetJob`
        """
        raise NotImplementedError("create_target_job not implemented for this driver")

    def resume_target_job(self, target, job):
        """
        Resume a suspended backup job on a target

        :param target: Backup target with the backup data
        :type  target: Instance of :class:`BackupTarget`

        :param job: Backup target job to resume
        :type  job: Instance of :class:`BackupTargetJob`

        :rtype: ``bool``
        """
        raise NotImplementedError("resume_target_job not implemented for this driver")

    def suspend_target_job(self, target, job):
        """
        Suspend a running backup job on a target

        :param target: Backup target with the backup data
        :type  target: Instance of :class:`BackupTarget`

        :param job: Backup target job to suspend
        :type  job: Instance of :class:`BackupTargetJob`

        :rtype: ``bool``
        """
        raise NotImplementedError("suspend_target_job not implemented for this driver")

    def cancel_target_job(self, job, ex_client=None, ex_target=None):
        """
        Cancel a backup job on a target

        :param job: Backup target job to cancel.  If it is ``None``
                    ex_client and ex_target must be set
        :type  job: Instance of :class:`BackupTargetJob` or ``None``

        :param ex_client: Client of the job to cancel.
                          Not necessary if job is specified.
                          DimensionData only has 1 job per client
        :type  ex_client: Instance of :class:`DimensionDataBackupClient`
                          or ``str``

        :param ex_target: Target to cancel a job from.
                          Not necessary if job is specified.
        :type  ex_target: Instance of :class:`BackupTarget` or ``str``

        :rtype: ``bool``
        """
        if job is None:
            if ex_client is None or ex_target is None:
                raise ValueError("Either job or ex_client and " "ex_target have to be set")
            server_id = self._target_to_target_address(ex_target)
            client_id = self._client_to_client_id(ex_client)
        else:
            server_id = job.target.address
            client_id = job.extra["clientId"]

        response = self.connection.request_with_orgId_api_1(
            "server/{}/backup/client/{}?cancelJob".format(server_id, client_id),
            method="GET",
        ).object
        response_code = findtext(response, "result", GENERAL_NS)
        return response_code in ["IN_PROGRESS", "SUCCESS"]

    def ex_get_target_by_id(self, id):
        """
        Get a target by server id

        :param id: The id of the target you want to get
        :type  id: ``str``

        :rtype: :class:`BackupTarget`
        """
        node = self.connection.request_with_orgId_api_2("server/server/%s" % id).object
        return self._to_target(node)

    def ex_add_client_to_target(
        self, target, client_type, storage_policy, schedule_policy, trigger, email
    ):
        """
        Add a client to a target

        :param target: Backup target with the backup data
        :type  target: Instance of :class:`BackupTarget` or ``str``

        :param client: Client to add to the target
        :type  client: Instance of :class:`DimensionDataBackupClientType`
                       or ``str``

        :param storage_policy: The storage policy for the client
        :type  storage_policy: Instance of
                               :class:`DimensionDataBackupStoragePolicy`
                               or ``str``

        :param schedule_policy: The schedule policy for the client
        :type  schedule_policy: Instance of
                                :class:`DimensionDataBackupSchedulePolicy`
                                or ``str``

        :param trigger: The notify trigger for the client
        :type  trigger: ``str``

        :param email: The notify email for the client
        :type  email: ``str``

        :rtype: ``bool``
        """
        server_id = self._target_to_target_address(target)

        backup_elm = ET.Element("NewBackupClient", {"xmlns": BACKUP_NS})
        if isinstance(client_type, DimensionDataBackupClientType):
            ET.SubElement(backup_elm, "type").text = client_type.type
        else:
            ET.SubElement(backup_elm, "type").text = client_type

        if isinstance(storage_policy, DimensionDataBackupStoragePolicy):
            ET.SubElement(backup_elm, "storagePolicyName").text = storage_policy.name
        else:
            ET.SubElement(backup_elm, "storagePolicyName").text = storage_policy

        if isinstance(schedule_policy, DimensionDataBackupSchedulePolicy):
            ET.SubElement(backup_elm, "schedulePolicyName").text = schedule_policy.name
        else:
            ET.SubElement(backup_elm, "schedulePolicyName").text = schedule_policy

        alerting_elm = ET.SubElement(backup_elm, "alerting")
        alerting_elm.set("trigger", trigger)
        ET.SubElement(alerting_elm, "emailAddress").text = email

        response = self.connection.request_with_orgId_api_1(
            "server/%s/backup/client" % (server_id),
            method="POST",
            data=ET.tostring(backup_elm),
        ).object
        response_code = findtext(response, "result", GENERAL_NS)
        return response_code in ["IN_PROGRESS", "SUCCESS"]

    def ex_remove_client_from_target(self, target, backup_client):
        """
        Removes a client from a backup target

        :param  target: The backup target to remove the client from
        :type   target: :class:`BackupTarget` or ``str``

        :param  backup_client: The backup client to remove
        :type   backup_client: :class:`DimensionDataBackupClient` or ``str``

        :rtype: ``bool``
        """
        server_id = self._target_to_target_address(target)
        client_id = self._client_to_client_id(backup_client)
        response = self.connection.request_with_orgId_api_1(
            "server/{}/backup/client/{}?disable".format(server_id, client_id), method="GET"
        ).object
        response_code = findtext(response, "result", GENERAL_NS)
        return response_code in ["IN_PROGRESS", "SUCCESS"]

    def ex_get_backup_details_for_target(self, target):
        """
        Returns a backup details object for a target

        :param  target: The backup target to get details for
        :type   target: :class:`BackupTarget` or ``str``

        :rtype: :class:`DimensionDataBackupDetails`
        """
        if not isinstance(target, BackupTarget):
            target = self.ex_get_target_by_id(target)
            if target is None:
                return
        response = self.connection.request_with_orgId_api_1(
            "server/%s/backup" % (target.address), method="GET"
        ).object
        return self._to_backup_details(response, target)

    def ex_list_available_client_types(self, target):
        """
        Returns a list of available backup client types

        :param  target: The backup target to list available types for
        :type   target: :class:`BackupTarget` or ``str``

        :rtype: ``list`` of :class:`DimensionDataBackupClientType`
        """
        server_id = self._target_to_target_address(target)
        response = self.connection.request_with_orgId_api_1(
            "server/%s/backup/client/type" % (server_id), method="GET"
        ).object
        return self._to_client_types(response)

    def ex_list_available_storage_policies(self, target):
        """
        Returns a list of available backup storage policies

        :param  target: The backup target to list available policies for
        :type   target: :class:`BackupTarget` or ``str``

        :rtype: ``list`` of :class:`DimensionDataBackupStoragePolicy`
        """
        server_id = self._target_to_target_address(target)
        response = self.connection.request_with_orgId_api_1(
            "server/%s/backup/client/storagePolicy" % (server_id), method="GET"
        ).object
        return self._to_storage_policies(response)

    def ex_list_available_schedule_policies(self, target):
        """
        Returns a list of available backup schedule policies

        :param  target: The backup target to list available policies for
        :type   target: :class:`BackupTarget` or ``str``

        :rtype: ``list`` of :class:`DimensionDataBackupSchedulePolicy`
        """
        server_id = self._target_to_target_address(target)
        response = self.connection.request_with_orgId_api_1(
            "server/%s/backup/client/schedulePolicy" % (server_id), method="GET"
        ).object
        return self._to_schedule_policies(response)

    def _to_storage_policies(self, object):
        elements = object.findall(fixxpath("storagePolicy", BACKUP_NS))

        return [self._to_storage_policy(el) for el in elements]

    def _to_storage_policy(self, element):
        return DimensionDataBackupStoragePolicy(
            retention_period=int(element.get("retentionPeriodInDays")),
            name=element.get("name"),
            secondary_location=element.get("secondaryLocation"),
        )

    def _to_schedule_policies(self, object):
        elements = object.findall(fixxpath("schedulePolicy", BACKUP_NS))

        return [self._to_schedule_policy(el) for el in elements]

    def _to_schedule_policy(self, element):
        return DimensionDataBackupSchedulePolicy(
            name=element.get("name"), description=element.get("description")
        )

    def _to_client_types(self, object):
        elements = object.findall(fixxpath("backupClientType", BACKUP_NS))

        return [self._to_client_type(el) for el in elements]

    def _to_client_type(self, element):
        description = element.get("description")
        if description is None:
            description = findtext(element, "description", BACKUP_NS)
        return DimensionDataBackupClientType(
            type=element.get("type"),
            description=description,
            is_file_system=bool(element.get("isFileSystem") == "true"),
        )

    def _to_backup_details(self, object, target):
        return DimensionDataBackupDetails(
            asset_id=object.get("assetId"),
            service_plan=object.get("servicePlan"),
            status=object.get("state"),
            clients=self._to_clients(object, target),
        )

    def _to_clients(self, object, target):
        elements = object.findall(fixxpath("backupClient", BACKUP_NS))

        return [self._to_client(el, target) for el in elements]

    def _to_client(self, element, target):
        client_id = element.get("id")
        return DimensionDataBackupClient(
            id=client_id,
            type=self._to_client_type(element),
            status=element.get("status"),
            schedule_policy=findtext(element, "schedulePolicyName", BACKUP_NS),
            storage_policy=findtext(element, "storagePolicyName", BACKUP_NS),
            download_url=findtext(element, "downloadUrl", BACKUP_NS),
            running_job=self._to_backup_job(element, target, client_id),
            alert=self._to_alert(element),
        )

    def _to_alert(self, element):
        alert = element.find(fixxpath("alerting", BACKUP_NS))
        if alert is not None:
            notify_list = [
                email_addr.text for email_addr in alert.findall(fixxpath("emailAddress", BACKUP_NS))
            ]
            return DimensionDataBackupClientAlert(
                trigger=element.get("trigger"), notify_list=notify_list
            )
        return None

    def _to_backup_job(self, element, target, client_id):
        running_job = element.find(fixxpath("runningJob", BACKUP_NS))
        if running_job is not None:
            return BackupTargetJob(
                id=running_job.get("id"),
                status=running_job.get("status"),
                progress=int(running_job.get("percentageComplete")),
                driver=self.connection.driver,
                target=target,
                extra={"clientId": client_id},
            )
        return None

    def _to_targets(self, object):
        node_elements = object.findall(fixxpath("server", TYPES_URN))

        return [self._to_target(el) for el in node_elements]

    def _to_target(self, element):
        backup = findall(element, "backup", TYPES_URN)
        if len(backup) == 0:
            return
        extra = {
            "description": findtext(element, "description", TYPES_URN),
            "sourceImageId": findtext(element, "sourceImageId", TYPES_URN),
            "datacenterId": element.get("datacenterId"),
            "deployedTime": findtext(element, "createTime", TYPES_URN),
            "servicePlan": backup[0].get("servicePlan"),
        }

        n = BackupTarget(
            id=backup[0].get("assetId"),
            name=findtext(element, "name", TYPES_URN),
            address=element.get("id"),
            driver=self.connection.driver,
            type=BackupTargetType.VIRTUAL,
            extra=extra,
        )
        return n

    @staticmethod
    def _client_to_client_id(backup_client):
        return dd_object_to_id(backup_client, DimensionDataBackupClient)

    @staticmethod
    def _target_to_target_address(target):
        return dd_object_to_id(target, BackupTarget, id_value="address")



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import hmac
import time
import uuid
import base64
import hashlib

from libcloud.utils.py3 import ET, b, u, urlquote
from libcloud.utils.xml import findtext
from libcloud.common.base import XmlResponse, ConnectionUserAndKey
from libcloud.common.types import MalformedResponseError

__all__ = [
    "AliyunXmlResponse",
    "AliyunRequestSigner",
    "AliyunRequestSignerAlgorithmV1_0",
    "SignedAliyunConnection",
    "AliyunConnection",
    "SIGNATURE_VERSION_1_0",
    "DEFAULT_SIGNATURE_VERSION",
]

SIGNATURE_VERSION_1_0 = "1.0"
DEFAULT_SIGNATURE_VERSION = SIGNATURE_VERSION_1_0


class AliyunXmlResponse(XmlResponse):
    namespace = None

    def success(self):
        return 200 <= self.status < 300

    def parse_body(self):
        """
        Each response from Aliyun contains a request id and a host id.
        The response body is in utf-8 encoding.
        """
        if len(self.body) == 0 and not self.parse_zero_length_body:
            return self.body

        try:
            parser = ET.XMLParser(encoding="utf-8")
            body = ET.XML(self.body.encode("utf-8"), parser=parser)
        except Exception:
            raise MalformedResponseError(
                "Failed to parse XML", body=self.body, driver=self.connection.driver
            )
        self.request_id = findtext(element=body, xpath="RequestId", namespace=self.namespace)
        self.host_id = findtext(element=body, xpath="HostId", namespace=self.namespace)
        return body

    def parse_error(self):
        """
        Parse error responses from Aliyun.
        """
        body = super().parse_error()
        code, message = self._parse_error_details(element=body)
        request_id = findtext(element=body, xpath="RequestId", namespace=self.namespace)
        host_id = findtext(element=body, xpath="HostId", namespace=self.namespace)
        error = {
            "code": code,
            "message": message,
            "request_id": request_id,
            "host_id": host_id,
        }
        return u(error)

    def _parse_error_details(self, element):
        """
        Parse error code and message from the provided error element.

        :return: ``tuple`` with two elements: (code, message)
        :rtype: ``tuple``
        """
        code = findtext(element=element, xpath="Code", namespace=self.namespace)
        message = findtext(element=element, xpath="Message", namespace=self.namespace)

        return (code, message)


class AliyunRequestSigner:
    """
    Class handles signing the outgoing Aliyun requests.
    """

    def __init__(self, access_key, access_secret, version):
        """
        :param access_key: Access key.
        :type access_key: ``str``

        :param access_secret: Access secret.
        :type access_secret: ``str``

        :param version: API version.
        :type version: ``str``
        """
        self.access_key = access_key
        self.access_secret = access_secret
        self.version = version

    def get_request_params(self, params, method="GET", path="/"):
        return params

    def get_request_headers(self, params, headers, method="GET", path="/"):
        return params, headers


class AliyunRequestSignerAlgorithmV1_0(AliyunRequestSigner):
    """Aliyun request signer using signature version 1.0."""

    def get_request_params(self, params, method="GET", path="/"):
        params["Format"] = "XML"
        params["Version"] = self.version
        params["AccessKeyId"] = self.access_key
        params["SignatureMethod"] = "HMAC-SHA1"
        params["SignatureVersion"] = SIGNATURE_VERSION_1_0
        params["SignatureNonce"] = _get_signature_nonce()
        # TODO: Support 'ResourceOwnerAccount'
        params["Timestamp"] = time.strftime("%Y-%m-%dT%H:%M:%SZ", time.gmtime())
        params["Signature"] = self._sign_request(params, method, path)
        return params

    def _sign_request(self, params, method, path):
        """
        Sign Aliyun requests parameters and get the signature.

        StringToSign = HTTPMethod + '&' +
                       percentEncode('/') + '&' +
                       percentEncode(CanonicalizedQueryString)
        """
        keys = list(params.keys())
        keys.sort()
        pairs = []
        for key in keys:
            pairs.append("{}={}".format(_percent_encode(key), _percent_encode(params[key])))
        qs = urlquote("&".join(pairs), safe="-_.~")
        string_to_sign = "&".join((method, urlquote(path, safe=""), qs))
        b64_hmac = base64.b64encode(
            hmac.new(
                b(self._get_access_secret()), b(string_to_sign), digestmod=hashlib.sha1
            ).digest()
        )

        return b64_hmac.decode("utf8")

    def _get_access_secret(self):
        return "%s&" % self.access_secret


class AliyunConnection(ConnectionUserAndKey):
    pass


class SignedAliyunConnection(AliyunConnection):
    api_version = None

    def __init__(
        self,
        user_id,
        key,
        secure=True,
        host=None,
        port=None,
        url=None,
        timeout=None,
        proxy_url=None,
        retry_delay=None,
        backoff=None,
        api_version=None,
        signature_version=DEFAULT_SIGNATURE_VERSION,
    ):
        super().__init__(
            user_id=user_id,
            key=key,
            secure=secure,
            host=host,
            port=port,
            url=url,
            timeout=timeout,
            proxy_url=proxy_url,
            retry_delay=retry_delay,
            backoff=backoff,
        )

        self.signature_version = str(signature_version)

        if self.signature_version == "1.0":
            signer_cls = AliyunRequestSignerAlgorithmV1_0
        else:
            raise ValueError("Unsupported signature_version: %s" % signature_version)

        if api_version is not None:
            self.api_version = str(api_version)
        else:
            if self.api_version is None:
                raise ValueError("Unsupported null api_version")

        self.signer = signer_cls(
            access_key=self.user_id, access_secret=self.key, version=self.api_version
        )

    def add_default_params(self, params):
        params = self.signer.get_request_params(params=params, method=self.method, path=self.action)
        return params


def _percent_encode(encode_str):
    """
    Encode string to utf8, quote for url and replace '+' with %20,
    '*' with %2A and keep '~' not converted.

    :param src_str: ``str`` in the same encoding with sys.stdin,
                    default to encoding cp936.
    :return: ``str`` represents the encoded result
    :rtype: ``str``
    """
    encoding = sys.stdin.encoding or "cp936"
    decoded = str(encode_str)

    if isinstance(encode_str, bytes):
        decoded = encode_str.decode(encoding)

    res = urlquote(decoded.encode("utf8"), "")
    res = res.replace("+", "%20")
    res = res.replace("*", "%2A")
    res = res.replace("%7E", "~")
    return res


def _get_signature_nonce():
    return str(uuid.uuid4())



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

try:
    import simplejson as json
except ImportError:
    import json

from libcloud.common.aws import AWSJsonResponse, SignedAWSConnection
from libcloud.container.base import Container, ContainerImage, ContainerDriver, ContainerCluster
from libcloud.container.types import ContainerState
from libcloud.container.utils.docker import RegistryClient

__all__ = ["ElasticContainerDriver"]


ECS_VERSION = "2014-11-13"
ECR_VERSION = "2015-09-21"
ECS_HOST = "ecs.%s.amazonaws.com"
ECR_HOST = "ecr.%s.amazonaws.com"
ROOT = "/"
ECS_TARGET_BASE = "AmazonEC2ContainerServiceV%s" % (ECS_VERSION.replace("-", ""))
ECR_TARGET_BASE = "AmazonEC2ContainerRegistry_V%s" % (ECR_VERSION.replace("-", ""))


class ECSJsonConnection(SignedAWSConnection):
    version = ECS_VERSION
    host = ECS_HOST
    responseCls = AWSJsonResponse
    service_name = "ecs"


class ECRJsonConnection(SignedAWSConnection):
    version = ECR_VERSION
    host = ECR_HOST
    responseCls = AWSJsonResponse
    service_name = "ecr"


class ElasticContainerDriver(ContainerDriver):
    name = "Amazon Elastic Container Service"
    website = "https://aws.amazon.com/ecs/details/"
    ecr_repository_host = "%s.dkr.ecr.%s.amazonaws.com"
    connectionCls = ECSJsonConnection
    ecrConnectionClass = ECRJsonConnection
    supports_clusters = False
    status_map = {"RUNNING": ContainerState.RUNNING}

    def __init__(self, access_id, secret, region):
        super().__init__(access_id, secret)
        self.region = region
        self.region_name = region
        self.connection.host = ECS_HOST % (region)

        # Setup another connection class for ECR
        conn_kwargs = self._ex_connection_class_kwargs()
        self.ecr_connection = self.ecrConnectionClass(access_id, secret, **conn_kwargs)
        self.ecr_connection.host = ECR_HOST % (region)
        self.ecr_connection.driver = self
        self.ecr_connection.connect()

    def _ex_connection_class_kwargs(self):
        return {"signature_version": "4"}

    def list_images(self, ex_repository_name):
        """
        List the images in an ECR repository

        :param  ex_repository_name: The name of the repository to check
            defaults to the default repository.
        :type   ex_repository_name: ``str``

        :return: a list of images
        :rtype: ``list`` of :class:`libcloud.container.base.ContainerImage`
        """
        request = {}
        request["repositoryName"] = ex_repository_name
        list_response = self.ecr_connection.request(
            ROOT,
            method="POST",
            data=json.dumps(request),
            headers=self._get_ecr_headers("ListImages"),
        ).object
        repository_id = self.ex_get_repository_id(ex_repository_name)
        host = self._get_ecr_host(repository_id)
        return self._to_images(list_response["imageIds"], host, ex_repository_name)

    def list_clusters(self):
        """
        Get a list of potential locations to deploy clusters into

        :param  location: The location to search in
        :type   location: :class:`libcloud.container.base.ClusterLocation`

        :rtype: ``list`` of :class:`libcloud.container.base.ContainerCluster`
        """
        listdata = self.connection.request(
            ROOT,
            method="POST",
            data=json.dumps({}),
            headers=self._get_headers("ListClusters"),
        ).object
        request = {"clusters": listdata["clusterArns"]}
        data = self.connection.request(
            ROOT,
            method="POST",
            data=json.dumps(request),
            headers=self._get_headers("DescribeClusters"),
        ).object
        return self._to_clusters(data)

    def create_cluster(self, name, location=None):
        """
        Create a container cluster

        :param  name: The name of the cluster
        :type   name: ``str``

        :param  location: The location to create the cluster in
        :type   location: :class:`libcloud.container.base.ClusterLocation`

        :rtype: :class:`libcloud.container.base.ContainerCluster`
        """
        request = {"clusterName": name}
        response = self.connection.request(
            ROOT,
            method="POST",
            data=json.dumps(request),
            headers=self._get_headers("CreateCluster"),
        ).object
        return self._to_cluster(response["cluster"])

    def destroy_cluster(self, cluster):
        """
        Delete a cluster

        :return: ``True`` if the destroy was successful, otherwise ``False``.
        :rtype: ``bool``
        """
        request = {"cluster": cluster.id}
        data = self.connection.request(
            ROOT,
            method="POST",
            data=json.dumps(request),
            headers=self._get_headers("DeleteCluster"),
        ).object
        return data["cluster"]["status"] == "INACTIVE"

    def list_containers(self, image=None, cluster=None):
        """
        List the deployed container images

        :param image: Filter to containers with a certain image
        :type  image: :class:`libcloud.container.base.ContainerImage`

        :param cluster: Filter to containers in a cluster
        :type  cluster: :class:`libcloud.container.base.ContainerCluster`

        :rtype: ``list`` of :class:`libcloud.container.base.Container`
        """
        request = {"cluster": "default"}
        if cluster is not None:
            request["cluster"] = cluster.id
        if image is not None:
            request["family"] = image.name
        list_response = self.connection.request(
            ROOT,
            method="POST",
            data=json.dumps(request),
            headers=self._get_headers("ListTasks"),
        ).object
        if len(list_response["taskArns"]) == 0:
            return []
        containers = self.ex_list_containers_for_task(list_response["taskArns"])
        return containers

    def deploy_container(
        self,
        name,
        image,
        cluster=None,
        parameters=None,
        start=True,
        ex_cpu=10,
        ex_memory=500,
        ex_container_port=None,
        ex_host_port=None,
    ):
        """
        Creates a task definition from a container image that can be run
        in a cluster.

        :param name: The name of the new container
        :type  name: ``str``

        :param image: The container image to deploy
        :type  image: :class:`libcloud.container.base.ContainerImage`

        :param cluster: The cluster to deploy to, None is default
        :type  cluster: :class:`libcloud.container.base.ContainerCluster`

        :param parameters: Container Image parameters
        :type  parameters: ``str``

        :param start: Start the container on deployment
        :type  start: ``bool``

        :rtype: :class:`libcloud.container.base.Container`
        """
        data = {}
        if ex_container_port is None and ex_host_port is None:
            port_maps = []
        else:
            port_maps = [{"containerPort": ex_container_port, "hostPort": ex_host_port}]
        data["containerDefinitions"] = [
            {
                "mountPoints": [],
                "name": name,
                "image": image.name,
                "cpu": ex_cpu,
                "environment": [],
                "memory": ex_memory,
                "portMappings": port_maps,
                "essential": True,
                "volumesFrom": [],
            }
        ]
        data["family"] = name
        response = self.connection.request(
            ROOT,
            method="POST",
            data=json.dumps(data),
            headers=self._get_headers("RegisterTaskDefinition"),
        ).object
        if start:
            return self.ex_start_task(response["taskDefinition"]["taskDefinitionArn"])[0]
        else:
            return Container(
                id=None,
                name=name,
                image=image,
                state=ContainerState.RUNNING,
                ip_addresses=[],
                extra={"taskDefinitionArn": response["taskDefinition"]["taskDefinitionArn"]},
                driver=self.connection.driver,
            )

    def get_container(self, id):
        """
        Get a container by ID

        :param id: The ID of the container to get
        :type  id: ``str``

        :rtype: :class:`libcloud.container.base.Container`
        """
        containers = self.ex_list_containers_for_task([id])
        return containers[0]

    def start_container(self, container, count=1):
        """
        Start a deployed task

        :param container: The container to start
        :type  container: :class:`libcloud.container.base.Container`

        :param count: Number of containers to start
        :type  count: ``int``

        :rtype: :class:`libcloud.container.base.Container`
        """
        return self.ex_start_task(container.extra["taskDefinitionArn"], count)

    def stop_container(self, container):
        """
        Stop a deployed container

        :param container: The container to stop
        :type  container: :class:`libcloud.container.base.Container`

        :rtype: :class:`libcloud.container.base.Container`
        """
        request = {"task": container.extra["taskArn"]}
        response = self.connection.request(
            ROOT,
            method="POST",
            data=json.dumps(request),
            headers=self._get_headers("StopTask"),
        ).object
        containers = []
        containers.extend(
            self._to_containers(response["task"], container.extra["taskDefinitionArn"])
        )
        return containers

    def restart_container(self, container):
        """
        Restart a deployed container

        :param container: The container to restart
        :type  container: :class:`libcloud.container.base.Container`

        :rtype: :class:`libcloud.container.base.Container`
        """
        self.stop_container(container)
        return self.start_container(container)

    def destroy_container(self, container):
        """
        Destroy a deployed container

        :param container: The container to destroy
        :type  container: :class:`libcloud.container.base.Container`

        :rtype: :class:`libcloud.container.base.Container`
        """
        return self.stop_container(container)

    def ex_start_task(self, task_arn, count=1):
        """
        Run a task definition and get the containers

        :param task_arn: The task ARN to Run
        :type  task_arn: ``str``

        :param count: The number of containers to start
        :type  count: ``int``

        :rtype: ``list`` of :class:`libcloud.container.base.Container`
        """
        request = None
        request = {"count": count, "taskDefinition": task_arn}
        response = self.connection.request(
            ROOT,
            method="POST",
            data=json.dumps(request),
            headers=self._get_headers("RunTask"),
        ).object
        containers = []
        for task in response["tasks"]:
            containers.extend(self._to_containers(task, task_arn))
        return containers

    def ex_list_containers_for_task(self, task_arns):
        """
        Get a list of containers by ID collection (ARN)

        :param task_arns: The list of ARNs
        :type  task_arns: ``list`` of ``str``

        :rtype: ``list`` of :class:`libcloud.container.base.Container`
        """
        describe_request = {"tasks": task_arns}
        descripe_response = self.connection.request(
            ROOT,
            method="POST",
            data=json.dumps(describe_request),
            headers=self._get_headers("DescribeTasks"),
        ).object
        containers = []
        for task in descripe_response["tasks"]:
            containers.extend(self._to_containers(task, task["taskDefinitionArn"]))
        return containers

    def ex_create_service(self, name, cluster, task_definition, desired_count=1):
        """
        Runs and maintains a desired number of tasks from a specified
        task definition. If the number of tasks running in a service
        drops below desired_count, Amazon ECS spawns another
        instantiation of the task in the specified cluster.

        :param  name: the name of the service
        :type   name: ``str``

        :param  cluster: The cluster to run the service on
        :type   cluster: :class:`libcloud.container.base.ContainerCluster`

        :param  task_definition: The task definition name or ARN for the
            service
        :type   task_definition: ``str``

        :param  desired_count: The desired number of tasks to be running
            at any one time
        :type   desired_count: ``int``

        :rtype: ``object`` The service object
        """
        request = {
            "serviceName": name,
            "taskDefinition": task_definition,
            "desiredCount": desired_count,
            "cluster": cluster.id,
        }
        response = self.connection.request(
            ROOT,
            method="POST",
            data=json.dumps(request),
            headers=self._get_headers("CreateService"),
        ).object
        return response["service"]

    def ex_list_service_arns(self, cluster=None):
        """
        List the services

        :param cluster: The cluster hosting the services
        :type  cluster: :class:`libcloud.container.base.ContainerCluster`

        :rtype: ``list`` of ``str``
        """
        request = {}
        if cluster is not None:
            request["cluster"] = cluster.id
        response = self.connection.request(
            ROOT,
            method="POST",
            data=json.dumps(request),
            headers=self._get_headers("ListServices"),
        ).object
        return response["serviceArns"]

    def ex_describe_service(self, service_arn):
        """
        Get the details of a service

        :param  cluster: The hosting cluster
        :type   cluster: :class:`libcloud.container.base.ContainerCluster`

        :param  service_arn: The service ARN to describe
        :type   service_arn: ``str``

        :return: The service object
        :rtype: ``object``
        """
        request = {"services": [service_arn]}
        response = self.connection.request(
            ROOT,
            method="POST",
            data=json.dumps(request),
            headers=self._get_headers("DescribeServices"),
        ).object
        return response["services"][0]

    def ex_destroy_service(self, service_arn):
        """
        Deletes a service

        :param  cluster: The target cluster
        :type   cluster: :class:`libcloud.container.base.ContainerCluster`

        :param  service_arn: The service ARN to destroy
        :type   service_arn: ``str``
        """
        request = {"service": service_arn}
        response = self.connection.request(
            ROOT,
            method="POST",
            data=json.dumps(request),
            headers=self._get_headers("DeleteService"),
        ).object
        return response["service"]

    def ex_get_registry_client(self, repository_name):
        """
        Get a client for an ECR repository

        :param  repository_name: The unique name of the repository
        :type   repository_name: ``str``

        :return: a docker registry API client
        :rtype: :class:`libcloud.container.utils.docker.RegistryClient`
        """
        repository_id = self.ex_get_repository_id(repository_name)
        token = self.ex_get_repository_token(repository_id)
        host = self._get_ecr_host(repository_id)
        return RegistryClient(host=host, username="AWS", password=token)

    def ex_get_repository_token(self, repository_id):
        """
        Get the authorization token (12 hour expiry) for a repository

        :param  repository_id: The ID of the repository
        :type   repository_id: ``str``

        :return: A token for login
        :rtype: ``str``
        """
        request = {"RegistryIds": [repository_id]}
        response = self.ecr_connection.request(
            ROOT,
            method="POST",
            data=json.dumps(request),
            headers=self._get_ecr_headers("GetAuthorizationToken"),
        ).object
        return response["authorizationData"][0]["authorizationToken"]

    def ex_get_repository_id(self, repository_name):
        """
        Get the ID of a repository

        :param  repository_name: The unique name of the repository
        :type   repository_name: ``str``

        :return: The repository ID
        :rtype: ``str``
        """
        request = {"repositoryNames": [repository_name]}
        list_response = self.ecr_connection.request(
            ROOT,
            method="POST",
            data=json.dumps(request),
            headers=self._get_ecr_headers("DescribeRepositories"),
        ).object
        repository_id = list_response["repositories"][0]["registryId"]
        return repository_id

    def _get_ecr_host(self, repository_id):
        return self.ecr_repository_host % (repository_id, self.region)

    def _get_headers(self, action):
        """
        Get the default headers for a request to the ECS API
        """
        return {
            "x-amz-target": "{}.{}".format(ECS_TARGET_BASE, action),
            "Content-Type": "application/x-amz-json-1.1",
        }

    def _get_ecr_headers(self, action):
        """
        Get the default headers for a request to the ECR API
        """
        return {
            "x-amz-target": "{}.{}".format(ECR_TARGET_BASE, action),
            "Content-Type": "application/x-amz-json-1.1",
        }

    def _to_clusters(self, data):
        clusters = []
        for cluster in data["clusters"]:
            clusters.append(self._to_cluster(cluster))
        return clusters

    def _to_cluster(self, data):
        return ContainerCluster(
            id=data["clusterArn"],
            name=data["clusterName"],
            driver=self.connection.driver,
        )

    def _to_containers(self, data, task_definition_arn):
        clusters = []
        for cluster in data["containers"]:
            clusters.append(self._to_container(cluster, task_definition_arn))
        return clusters

    def _to_container(self, data, task_definition_arn):
        return Container(
            id=data["containerArn"],
            name=data["name"],
            image=ContainerImage(
                id=None,
                name=data["name"],
                path=None,
                version=None,
                driver=self.connection.driver,
            ),
            ip_addresses=None,
            state=self.status_map.get(data["lastStatus"], None),
            extra={
                "taskArn": data["taskArn"],
                "taskDefinitionArn": task_definition_arn,
            },
            driver=self.connection.driver,
        )

    def _to_images(self, data, host, repository_name):
        images = []
        for image in data:
            images.append(self._to_image(image, host, repository_name))
        return images

    def _to_image(self, data, host, repository_name):
        path = "{}/{}:{}".format(host, repository_name, data["imageTag"])
        return ContainerImage(
            id=None,
            name=path,
            path=path,
            version=data["imageTag"],
            driver=self.connection.driver,
        )
